try:
    from sympy.printing.ccode import CCodePrinter
    from sympy.printing.precedence import precedence
    import sympy
except ImportError:
    sympy = None
    CCodePrinter = object
from ...optimiser import symbolic_eval

def boolean_printer(origop, newop, start=''):
    def f(self, expr):
        PREC = precedence(expr)
        return start + newop.join(self.parenthesize(a, PREC) for a in expr.args)
    f.__name__ = '_print_' + origop
    return f


class NewCCodePrinter(CCodePrinter):
    _print_And = boolean_printer('And', '&&')
    _print_Or = boolean_printer('Or', '||')
    _print_Not = boolean_printer('Not', '', '!')

__all__ = ['rewrite_to_c_expression', 'sympy_rewrite', 'rewrite_pow', 'floatify_numbers']

if sympy is not None:
    pow = sympy.Function('pow')
else:
    pow = None

def rewrite_pow(e):
    if not len(e.args):
        return e
    newargs = tuple(rewrite_pow(a) for a in e.args)
    if isinstance(e, sympy.Pow) and e.args[1] != -1:
        return pow(*newargs)
    else:
        return e.new(*newargs)

def floatify_numbers(e):
    if not len(e.args):
        if e.is_number:
            return sympy.Number(float(e))
        return e
    newargs = tuple(floatify_numbers(a) for a in e.args)
    return e.new(*newargs)

def make_sympy_expressions(eqs):
    exprs = {}
    for name in eqs._diffeq_names + eqs._eq_names:
        exprs[name] = symbolic_eval(eqs._string[name])
    return exprs

def sympy_rewrite(s, rewriters=None):
    if rewriters is not None:
        if callable(rewriters):
            rewriters = [rewriters]
    else:
        rewriters = []
    expr = symbolic_eval(s)
    if not hasattr(expr, 'args'):
        return str(expr)
    for f in rewriters:
        expr = f(expr)
    return str(expr)

def rewrite_to_c_expression(s):
    if sympy is None:
        raise ImportError('sympy package required for code generation.')
    e = symbolic_eval(s)
    return NewCCodePrinter().doprint(e)

def generate_c_expressions(eqs):
    exprs = make_sympy_expressions(eqs)
    cexprs = {}
    for name, expr in exprs.iteritems():
        cexprs[name] = str(rewrite_pow(expr))
    return cexprs

if __name__ == '__main__':
    from brian import *
    if True:
        s = '-V**2/(10*0.001)'
        print sympy_rewrite(s)
        print sympy_rewrite(s, rewrite_pow)
        print sympy_rewrite(s, floatify_numbers)
        print sympy_rewrite(s, [rewrite_pow, floatify_numbers])
    if False:
        area = 20000 * umetre ** 2
        Cm = (1 * ufarad * cm ** -2) * area
        gl = (5e-5 * siemens * cm ** -2) * area
        El = -60 * mV
        EK = -90 * mV
        ENa = 50 * mV
        g_na = (100 * msiemens * cm ** -2) * area
        g_kd = (30 * msiemens * cm ** -2) * area
        VT = -63 * mV
        # Time constants
        taue = 5 * ms
        taui = 10 * ms
        # Reversal potentials
        Ee = 0 * mV
        Ei = -80 * mV
        we = 6 * nS # excitatory synaptic weight (voltage)
        wi = 67 * nS # inhibitory synaptic weight
        # The model
        eqs = Equations('''
        dv/dt = (gl*(El-v)+ge*(Ee-v)+gi*(Ei-v)-g_na*(m*m*m)*h*(v-ENa)-g_kd*(n*n*n*n)*(v-EK))/Cm : volt 
        dm/dt = alpham*(1-m)-betam*m : 1
        dn/dt = alphan*(1-n)-betan*n : 1
        dh/dt = alphah*(1-h)-betah*h : 1
        dge/dt = -ge*(1./taue) : siemens
        dgi/dt = -gi*(1./taui) : siemens
        alpham = 0.32*(mV**-1)*(13*mV-v+VT)/(exp((13*mV-v+VT)/(4*mV))-1.)/ms : Hz
        betam = 0.28*(mV**-1)*(v-VT-40*mV)/(exp((v-VT-40*mV)/(5*mV))-1)/ms : Hz
        alphah = 0.128*exp((17*mV-v+VT)/(18*mV))/ms : Hz
        betah = 4./(1+exp((40*mV-v+VT)/(5*mV)))/ms : Hz
        alphan = 0.032*(mV**-1)*(15*mV-v+VT)/(exp((15*mV-v+VT)/(5*mV))-1.)/ms : Hz
        betan = .5*exp((10*mV-v+VT)/(40*mV))/ms : Hz
        ''')
        cexprs = generate_c_expressions(eqs)
        for k, v in cexprs.iteritems():
            print k, ':', v
    if False:
        taum = 20 * msecond
        taue = 5 * msecond
        taui = 10 * msecond
        Ee = (0. + 60.) * mvolt
        Ei = (-80. + 60.) * mvolt
        eqs = Equations('''
        dv/dt = (-v+ge*(Ee-v)+gi*(Ei-v))*(1./taum) : volt
        dge/dt = -ge*(1./taue) : 1
        dgi/dt = -gi*(1./taui) : 1 
        ''')
        cexprs = generate_c_expressions(eqs)
        for k, v in cexprs.iteritems():
            print k, ':', v
    if False:
        x = sympy.Symbol('x')
        y = x ** 2 + 3 * x
        print y
        print rewrite_pow(y)



from brian import *
from brian.library.random_processes import *

# General variables
N = 100
RN_idx = 10 # the reinforced neuron
connP = 0.02 # Connection probability for reduced noise connections
connP_EE_scale = 1.0 # Scaling factor for EE connection
connP_IE_scale = 1.2 # Scaling factor for IE connection

# Parameters for conductance based Leaky Integrate and Fire Neuron

Cm = 0.3 * nfarad # Membrane capacitance
Rm = 100 * Mohm # Membrane resistance
# Time constants
taus = 5 * msecond # Excitatory/Inhibitory synapse time constant
# Potential settings
Ee = 0 * mV # Excitatory synapse reversal potential 
Ei = -75 * mV # Inhibitory synapse reversal potential
Vthresh = -59 * mV # Threshold Potential
Vresting = -70 * mV # Resting Potential (V(0))
Vreset = -70 * mvolt # Reset potential
# Delays
Trefrac = 5 * msecond # Refractory period
Tdelay = 1 * msecond # Synaptic time delay

# STDP variables
tau_pre = 30 * msecond
tau_post = 30 * msecond
Ap = 0.01
Am = -Ap * tau_pre / tau_post * 1.05 # approx -0.01

# RM-STDP specific variables

TauElig = 0.4 * second # Tau for eligibility trace
ArPos = 1.379 # Alpha plus for reward kernel
ArNeg = 0.27 # Alpha neg for reward kernel
TauRPos = 0.2 * second # Tau plus for reward kernel
TauRNeg = 1.0 * second # Tau neg for reward kernel
TDelay = 0.2 * second # Reward delay
last_rn_spike = 0.0 * second # Time of previous reinforced neuron spike
current_rn_spike = 0.0 * second # Time of most recent reinforced neuron spike
current_reward = 0.0 # Global variable for reward signal

# some arrays for plotting an eligibility trace
f_times = []
f_values = []


# O-U Noise equations

eqs_Enoise = OrnsteinUhlenbeck('Enoise', mu=0.012 * usiemens, sigma=0.003 * usiemens, tau=2.7 * ms)
eqs_Inoise = OrnsteinUhlenbeck('Inoise', mu=0.057 * usiemens, sigma=0.0066 * usiemens, tau=10.5 * ms)

# Equations for eligibility trace
# f is the eligibility kernel function
# c is the eligibility trace which is
# (STDP wt change * f)

elig_eqs = Equations('''
last_time:second
tc_elig = (t - last_time) *(1.0/(TauElig + TDelay)):1
f = 0.5 * tc_elig * exp(1.0 - tc_elig):1
c:1
''')

# Equations

eqs_full = Equations('''
dV/dt = (V - Vresting)*(1./(Cm*Rm)) + (ge*(V - Ee))*(1./Cm) + (gi*(V - Ei))*(1./Cm) - (0.2*Enoise*(V - Ee))*(1./Cm) - (0.2*Inoise*(V - Ei))*(1./Cm): volt
dge/dt = -ge*(1./taus) : siemens
dgi/dt = -gi*(1./taus) : siemens
''')


eqs_main = eqs_full + eqs_Enoise + eqs_Inoise + elig_eqs

# Main neuron groups

Pe = NeuronGroup(N / 2, model=eqs_main, threshold=Vthresh, reset=Vreset, refractory=Trefrac, implicit=True)
Pe.V = Vresting
Pe.ge = 0 * mV
Pe.gi = 0 * mV
Pe.last_time = 0.0 * second

Pi = NeuronGroup(N / 2, model=eqs_main, threshold=Vthresh, reset=Vreset, refractory=Trefrac, implicit=True)
Pi.V = Vresting
Pi.ge = 0 * mV
Pi.gi = 0 * mV
Pi.last_time = 0.0 * second

# Synapse weight equation parameters

Wscale = 0.8
WExcScale = 1.0
WInhScale = 1.4
taum = Rm * Cm

we = ((Vthresh - Vresting) * WExcScale * Wscale) / ((Ee - Vresting) * Rm * taus / (taum - taus) * (((taus / taum) ** (taus / (taum - taus))) - ((taus / taum) ** (taum / (taum - taus)))))
wi = ((Vthresh - Vresting) * WInhScale * Wscale) / ((Vresting - Ei) * Rm * taus / (taum - taus) * (((taus / taum) ** (taus / (taum - taus))) - ((taus / taum) ** (taum / (taum - taus)))))

# Maximum weight allowable during STDP is set to 2 * initial excitatory weight

wMax = we * 2.0

# Make a neuron subgroup which is the reinforced neuron

RN = Pe[RN_idx]

Cee = Connection(Pe, Pe, 'ge', weight=we, sparseness=connP * connP_EE_scale, delay=Tdelay)
Cie = Connection(Pe, Pe, 'gi', weight=wi, sparseness=connP * connP_IE_scale, delay=Tdelay)


# Custom network operations to collect the eligibility
# trace for reinforced neuron for plottingNameError: global name 'TauElig' is not defined

@network_operation
def save_elig():
    f_times.append(defaultclock.t)
    f_values.append(Pe.f[RN_idx])


M = SpikeMonitor(Pe)
run(200 * msecond)
raster_plot(M)

# Create plot of elig values

figure()
plot(elig_times, elig, 'k')
xlabel("Time/ms")
ylabel("Elig")
suptitle('Eligibility Trace', fontsize=12)





from brian import *
from brian.experimental.neuromorphic import *

defaultclock.dt = 1*ms
runtime = 100*ms

addr, timestamps = load_AER('./dummy.aedat')

N = max(addr)
group = NewSpikeGeneratorGroup(N, (addr, timestamps), time_unit = 1*usecond)

M = SpikeMonitor(group)

run(100*ms)

raster_plot(M)
show()



from brian import *
from numpy import random
from scipy import sparse,linalg

N_PY = 100
alpha_x = 1.0

t_final = 0.01*second

sim_clock = Clock(dt=0.05*ms)

PY = NeuronGroup(N=N_PY,
                 model='V:1',clock=sim_clock,
                 #threshold=EmpiricalThreshold(state='Vs',threshold=10,refractory=3*ms,clock=sim_clock),
                 threshold=10,
                 order=2,freeze=True)

CPP = zeros((N_PY,N_PY))
CPP[20,50] = 1.

CPP = sparse.lil_matrix(CPP)

C = Connection(PY,PY,'V',weight=alpha_x*CPP)

run(t_final)



# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
'''
Logging information for Brian
'''

import logging
import sys

__all__ = [ 'log_warn', 'log_info', 'log_debug',
           'log_level_error', 'log_level_warn', 'log_level_info', 'log_level_debug' ]

console = logging.StreamHandler(sys.stderr)
formatter = logging.Formatter('%(name)-18s: %(levelname)-8s %(message)s')
console.setFormatter(formatter)
logging.getLogger('brian').addHandler(console)

#brian_loggers = {}

get_log = logging.getLogger
#def get_log(logname):
#    if logname in brian_loggers:
#        return brian_loggers[logname]
#    newlog = logging.getLogger(logname)
#    #newlog.addHandler(console)
#    brian_loggers[logname] = newlog
#    return newlog

def log_warn(logname, message):
    get_log(logname).warn(message)

def log_info(logname, message):
    get_log(logname).info(message)

def log_debug(logname, message):
    get_log(logname).debug(message)

def log_level_error():
    '''Shows log messages only of level ERROR or higher.
    '''
    logging.getLogger('brian').setLevel(logging.ERROR)

def log_level_warn():
    '''Shows log messages only of level WARNING or higher (including ERROR level).
    '''
    logging.getLogger('brian').setLevel(logging.WARN)

def log_level_info():
    '''Shows log messages only of level INFO or higher (including WARNING and ERROR levels).
    '''
    logging.getLogger('brian').setLevel(logging.INFO)

def log_level_debug():
    '''Shows log messages only of level DEBUG or higher (including INFO, WARNING and ERROR levels).
    '''
    logging.getLogger('brian').setLevel(logging.DEBUG)

if __name__ == '__main__':
    #log_level_info()
    log_warn('brian.fish', 'Warning')
    log_warn('brian.monkey', 'Ook')
    log_info('brian.moose', 'Mook')
    log_debug('brian.goat', 'Bah')



from brian import *
#G = NeuronGroup(1, 'v:1\nw:1', threshold='(v>1)&(w>1)')
#M = SpikeMonitor(G)
#G.v = 2
#run(1*ms)
#print M.spikes
#G.w = 2
#run(1*ms)
#print M.spikes

eqs = '''
dV/dt = dV_dt: 1
dV_dt = V/(3*ms) : 1/second
thresh = (V>1)&(dV_dt>1/second) : 1
'''

G = NeuronGroup(1, eqs, threshold=EmpiricalThreshold(0, 1 * ms, 'thresh'))
G.V = 0.1
M = MultiStateMonitor(G, ('V', 'dV_dt', 'thresh'), record=True)

run(10 * ms)

subplot(311)
M['V'].plot()
subplot(312)
M['dV_dt'].plot()
subplot(313)
M['thresh'].plot()
show()



'''
Poisson input to an IF model
'''
from brian import *

PG = PoissonGroup(1, lambda t:200 * Hz * (1 + cos(2 * pi * t * 50 * Hz)))
IF = NeuronGroup(1, model='dv/dt=-v/(10*ms) : volt', reset=0 * volt, threshold=10 * mV)

C = Connection(PG, IF, 'v', weight=3 * mV)

MS = SpikeMonitor(PG, True)
Mv = StateMonitor(IF, 'v', record=True)
rates = StateMonitor(PG, 'rate', record=True)

run(100 * ms)

subplot(211)
plot(rates.times / ms, rates[0] / Hz)
subplot(212)
plot(Mv.times / ms, Mv[0] / mV)

show()



from brian import *
set_global_preferences(usenewbrianhears=True)
from brian.hears import *

x = Sound(randn(1000, 2), samplerate=44100*Hz)

gfb = GammatoneFilterbank(x, [1*kHz, 2*kHz])

ff = FunctionFilterbank(gfb, lambda input:clip(input, 0, Inf))

eqs = '''
dv/dt = (I-v)/(1*ms) : 1
I : 1
'''

G = FilterbankGroup(ff, 'I', eqs)
M = StateMonitor(G, 'v', record=True)

run(x.duration)

M.plot()
show()



from codegen import *
from rewriting import *

__all__ = ['CCodeGenerator']


class CCodeGenerator(CodeGenerator):
    def __init__(self, dtype='double', sympy_rewrite=True):
        CodeGenerator.__init__(self, sympy_rewrite=sympy_rewrite)
        self._dtype = dtype

    def vartype(self):
        return self._dtype

    def initialisation(self, eqs):
        vartype = self.vartype()
        code = ''
        for j, name in enumerate(eqs._diffeq_names):
            code += vartype + ' *' + name + '__Sbase = _S+' + str(j) + '*num_neurons;\n'
        return code

    def generate(self, eqs, scheme):
        vartype = self.vartype()
        code = self.initialisation(eqs)
        code += 'for(int _i=0;_i<num_neurons;_i++){\n'
        for j, name in enumerate(eqs._diffeq_names):
            code += '    ' + vartype + ' &' + name + ' = *' + name + '__Sbase++;\n'
        for line in self.scheme(eqs, scheme).split('\n'):
            line = line.strip()
            if line:
                code += '    ' + line + '\n'
        code += '}\n'
        code += self.finalisation(eqs)
        return code

    def single_statement(self, expr):
        return CodeGenerator.single_statement(self, expr) + ';'

    def single_expr(self, expr):
        return rewrite_to_c_expression(CodeGenerator.single_expr(self, expr.strip())).strip()



import os, shutil
# remove old docs
os.chdir('../../../')
shutil.rmtree('docs')
os.mkdir('docs')
# Generate new docs
os.chdir('docs_sphinx') # work from docs_sphinx/ directory until we make the big change
# normally use build_html.py instead, faster
os.system('sphinx-build -a -E -b html . ../docs') # clean build



import brian_no_units
from brian import *
import time
import pickle
from operator import itemgetter
from cuba_runopts import *

set_global_preferences(useweave=True)

def cubanetwork(N):
    Vr = -49 * mV
    invtauv = 1 / (20 * ms)
    invtauge = 1 / (5 * ms)
    invtaugi = 1 / (10 * ms)
    eqs = '''
    dv/dt = (ge+gi-(v-Vr))*invtauv : volt
    dge/dt = -ge*invtauge : volt
    dgi/dt = -gi*invtaugi : volt
    '''

    Ne = int(N * 0.8)
    Ni = N - Ne

    P = NeuronGroup(N, model=eqs,
                  threshold= -50 * mV, reset= -60 * mV,
                  compile=True, freeze=True)
    P.v = -60 * mV + 10 * mV * rand(len(P))
    Pe = P.subgroup(Ne)
    Pi = P.subgroup(Ni)

    Ce = Connection(Pe, P, 'ge')
    Ci = Connection(Pi, P, 'gi')
    Ce.connect_random(Pe, P, 0.02, weight=1.62 * mV)
    Ci.connect_random(Pi, P, 0.02, weight= -9 * mV)

    net = Network(P, Ce, Ci)
    net.run(0 * ms)

    return P, Pe, Pi, Ce, Ci, net

def cuba(P, Pe, Pi, Ce, Ci, net):
    reinit_default_clock()
    P.v = -60 * mV + 10 * mV * rand(len(P))
    Ce.__init__(Pe, P, 'ge')
    Ci.__init__(Pi, P, 'gi')
    Ce.connect_random(Pe, P, 0.02, weight=1.62 * mV)
    Ci.connect_random(Pi, P, 0.02, weight= -9 * mV)
    tstart = time.time()
    net.run(duration)
    tend = time.time()
    return tend - tstart, 0

def cuba_average(N, repeats, best):
    cubadata = cubanetwork(N)
    cubaruns = [cuba(*cubadata) for i in range(repeats)]
    print 'Finished run', N
    cubaruns.sort(key=itemgetter(0))
    cubaruns = cubaruns[:best]
    t, ns = zip(*cubaruns)
    t = array(t)
    ns = array(ns)
    return N, mean(t), mean(ns), cubaruns

def cuba_runs(Nvals, repeats, best):
    return [cuba_average(N, repeats, best) for N in Nvals]

def do_runs(fname='brian_cuba_results_allopts.pkl'):
    ca = cuba_runs(Nvals, repeats, best)
    output = open(fname, 'wb')
    pickle.dump(ca, output, -1)
    output.close()
    return ca

if __name__ == '__main__':
    print do_runs()
#    cubadata = cubanetwork(32000)
#    print cuba(*cubadata)
#    def f():
#        cubadata = cubanetwork(4000)
#        print cuba(*cubadata)
#    f()
#    import cProfile as profile
#    profile.run('f()')
#    import hotshot, hotshot.stats
#    prof = hotshot.Profile('cuba_allopts.prof')
#    prof.runcall(f)
#    stats = hotshot.stats.load("cuba_allopts.prof")
#    stats.strip_dirs()
#    stats.sort_stats('cumulative','time','calls')
#    stats.print_stats()



import os, brian, datetime
from update_release_date import setreleasedate
setreleasedate()
pathname, filename = os.path.split(__file__)
os.chdir(pathname)
os.chdir('../../../.') # work from Brian's root
os.system('del MANIFEST')
os.system('setup.py bdist_wininst')
os.system('setup.py sdist --formats=gztar,zip')
os.chdir('dist')
bname = 'brian-' + brian.__version__
bdate = str(datetime.date.today())
def exec_command(cmd):
    print 'Executing command:', cmd
    os.system(cmd)
for ext in ['tar.gz', 'zip', 'win32.exe']:
    if os.path.exists(bname + '-' + bdate + '.' + ext):
        exec_command('del %s-%s.%s' % (bname, bdate, ext))
    exec_command('rename %s.%s %s-%s.%s' % (bname, ext, bname, bdate, ext))



'''
A model with adaptive threshold (increases with each spike)
'''
from brian import *

eqs = '''
dv/dt = -v/(10*ms) : volt
dvt/dt = (10*mV-vt)/(15*ms) : volt
'''

reset = '''
v=0*mV
vt+=3*mV
'''

IF = NeuronGroup(1, model=eqs, reset=reset, threshold='v>vt')
IF.rest()
PG = PoissonGroup(1, 500 * Hz)

C = Connection(PG, IF, 'v', weight=3 * mV)

Mv = StateMonitor(IF, 'v', record=True)
Mvt = StateMonitor(IF, 'vt', record=True)

run(100 * ms)

plot(Mv.times / ms, Mv[0] / mV)
plot(Mvt.times / ms, Mvt[0] / mV)

show()



from numpy import *
import pycuda
import pycuda.autoinit as autoinit
import pycuda.driver as drv
from pycuda import gpuarray
import time

N = 100000000
steps = 10
use_pagelocked = False
do_get = True
do_set = False

getset_factor = float(int(do_get) + int(do_set))

a_cpu = zeros(N)
a_gpu = pycuda.gpuarray.to_gpu(a_cpu)

start = time.time()

for _ in xrange(steps):
    if do_get: a_gpu.set(a_cpu)
    if do_set: a_gpu.get(a_cpu)

duration = time.time() - start

print 'Array size', N
print 'Pagelocked', use_pagelocked
print 'Number of steps', steps
print 'Download', do_get, 'upload', do_set
print 'Total memory copied (GB)', N * 8. * steps * getset_factor / 1024. ** 3
print 'Time taken (sec)', duration
print 'Transfer rate (GB/sec)', N * 8. * steps * getset_factor / 1024. ** 3 / duration



# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
'''
Threshold mechanisms
'''

__all__ = ['Threshold', 'FunThreshold', 'VariableThreshold', 'NoThreshold',
          'EmpiricalThreshold', 'SimpleFunThreshold', 'PoissonThreshold',
          'HomogeneousPoissonThreshold', 'StringThreshold']

from numpy import where, array, zeros, Inf
from units import *
from itertools import count
from clock import guess_clock, get_default_clock, reinit_default_clock
from random import sample # Python standard random module (sample is different)
from scipy import random
from numpy import clip
import bisect
from scipy import weave
from globalprefs import *
import warnings
from utils.approximatecomparisons import is_approx_equal
from log import *
import inspect
from inspection import *
import re
import numpy
CThreshold = PythonThreshold = None

def select_threshold(expr, eqs, level=0):
    '''
    Automatically selects the appropriate Threshold object from a string.
    
    Matches the following patterns:
    
    var_name > or >= const : Threshold
    var_name > or >= var_name : VariableThreshold
    others : StringThreshold
    '''
    global CThreshold, PythonThreshold
    use_codegen = get_global_preference('usecodegen') and get_global_preference('usecodegenthreshold')
    use_weave = get_global_preference('useweave') and get_global_preference('usecodegenweave')
    if use_codegen:
        if CThreshold is None:
            from experimental.codegen.threshold import CThreshold, PythonThreshold
        if use_weave:
            log_warn('brian.threshold', 'Using codegen CThreshold')
            return CThreshold(expr, level=level + 1)
        else:
            log_warn('brian.threshold', 'Using codegen PythonThreshold')
            return PythonThreshold(expr, level=level + 1)
    # plan:
    # - see if it matches A > B or A >= B, if not select StringThreshold
    # - check if A, B both match diffeq variable names, and if so
    #   select VariableThreshold
    # - check that A is a variable name, if not select StringThreshold
    # - extract all the identifiers from B, and if none of them are
    #   callable, assume it is a constant, try to eval it and then use
    #   Threshold. If not, or if eval fails, use StringThreshold.
    # This misses the case of e.g. V>10*mV*exp(1) because exp will be
    # callable, but in general a callable means that it could be
    # non-constant.
    expr = expr.strip()
    eqs.prepare()
    ns = namespace(expr, level=level + 1)
    s = re.search(r'^\s*(\w+)\s*>=?(.+)', expr)
    if not s:
        return StringThreshold(expr, level=level + 1)
    A = s.group(1)
    B = s.group(2).strip()
    if A not in eqs._diffeq_names:
        return StringThreshold(expr, level=level + 1)
    if B in eqs._diffeq_names:
        return VariableThreshold(B, A)
    try:
        vars = get_identifiers(B)
    except SyntaxError:
        return StringThreshold(expr, level=level + 1)
    all_vars = eqs._eq_names + eqs._diffeq_names + eqs._alias.keys() + ['t']
    for v in vars:
        if v not in ns or v in all_vars or callable(ns[v]):
            return StringThreshold(expr, level=level + 1)
    try:
        val = eval(B, ns)
    except:
        return StringThreshold(expr, level=level + 1)
    return Threshold(val, A)


class Threshold(object):
    '''
    All neurons with a specified state variable above a fixed value fire a spike.
    
    **Initialised as:** ::
    
        Threshold([threshold=1*mV[,state=0])
    
    with arguments:
    
    ``threshold``
        The value above which a neuron will fire.
    ``state``
        The state variable which is checked.
    
    **Compilation**
    
    Note that if the global variable ``useweave`` is set to ``True``
    then this function will use a ``C++`` accelerated version which
    runs approximately 3x faster.
    '''

    def __init__(self, threshold=1 * mvolt, state=0):
        self.threshold = threshold
        self.state = state
        self._useaccel = get_global_preference('useweave')
        self._cpp_compiler = get_global_preference('weavecompiler')
        self._extra_compile_args = ['-O3']
        if self._cpp_compiler == 'gcc':
            self._extra_compile_args += get_global_preference('gcc_options') # ['-march=native', '-ffast-math']

    def __call__(self, P):
        '''
        Checks the threshold condition and returns spike times.
        P is the neuron group.

        Note the accelerated version runs 3x faster.
        '''
        if self._useaccel:
            spikes = P._spikesarray
            V = P.state_(self.state)
            Vt = float(self.threshold)
            N = int(len(P))
            code = """
                    int numspikes=0;
                    for(int i=0;i<N;i++)
                        if(V(i)>Vt)
                            spikes(numspikes++) = i;
                    return_val = numspikes;
                    """
            # WEAVE NOTE: set the environment variable USER if your username has a space
            # in it, say set USER=DanGoodman if your username is Dan Goodman, this is
            # because weave uses this to create file names, but doesn't correctly send these
            # values to the compiler, causing problems.
            numspikes = weave.inline(code, ['spikes', 'V', 'Vt', 'N'],
                                     compiler=self._cpp_compiler,
                                     type_converters=weave.converters.blitz,
                                     extra_compile_args=self._extra_compile_args)
            # WEAVE NOTE: setting verbose=True in the weave.inline function may help in
            # finding errors.
            return spikes[0:numspikes]
        else:
            return ((P.state_(self.state) > self.threshold).nonzero())[0]

    def __repr__(self):
        return 'Threshold mechanism with value=' + str(self.threshold) + " acting on state " + str(self.state)


class StringThreshold(Threshold):
    '''
    A threshold specified by a string expression.
    
    Initialised with arguments:
    
    ``expr``
        The expression used to test whether a neuron has fired a spike.
        Should be a single statement that returns a value. For example,
        ``'V>50*mV'`` or ``'V>Vt'``.
    ``level``
        How many levels up in the calling sequence to look for
        names in the namespace. Usually 0 for user code.
    '''
    def __init__(self, expr, level=0):
        self._namespace, unknowns = namespace(expr, level=level + 1, return_unknowns=True)
        self._vars = unknowns
        self._expr = expr
        self._code = compile(expr, "StringThreshold", "eval")
        class Replacer(object):
            def __init__(self, func, n):
                self.n = n
                self.func = func
            def __call__(self):
                return self.func(self.n)
        self._Replacer = Replacer

    def __call__(self, P):
        for var in self._vars: # couldn't we do this just once?
            self._namespace[var] = P.state(var)
        self._namespace['rand'] = self._Replacer(numpy.random.rand, len(P))
        self._namespace['randn'] = self._Replacer(numpy.random.randn, len(P))
        return eval(self._code, self._namespace).nonzero()[0]

    def __repr__(self):
        return "String threshold"


class NoThreshold(Threshold):
    '''
    No thresholding mechanism.
    
    **Initialised as:** ::
    
        NoThreshold()
    '''
    def __init__(self):
        pass

    def __call__(self, P):
        return []

    def __repr__(self):
        return "No Threshold"


class FunThreshold(Threshold):
    '''
    Threshold mechanism with a user-specified function.
    
    **Initialised as:** ::
    
        FunThreshold(thresholdfun)
    
    where ``thresholdfun`` is a function with one argument,
    the 2d state value array, where each row is an array of
    values for one state, of length N for N the number of
    neurons in the group. For efficiency, data are numpy
    arrays and there is no unit checking.
    
    Note: if you only need to consider one state variable,
    use the :class:`SimpleFunThreshold` object instead.
    '''
    def __init__(self, thresholdfun):
        self.thresholdfun = thresholdfun # Threshold function

    def __call__(self, P):
        '''
        Checks the threshold condition and returns spike times.
        P is the neuron group.
        '''
        spikes = (self.thresholdfun(*P._S).nonzero())[0]
        return spikes

    def __repr__(self):
        return 'Functional threshold mechanism'


class SimpleFunThreshold(FunThreshold):
    '''
    Threshold mechanism with a user-specified function.
    
    **Initialised as:** ::
    
        FunThreshold(thresholdfun[,state=0])
    
    with arguments:
    
    ``thresholdfun``
        A function with one argument, the array of values for
        the specified state variable. For efficiency, this is
        a numpy array, and there is no unit checking.
    ``state``
        The name or number of the state variable to pass to
        the threshold function.
    
    **Sample usage:** ::
    
        FunThreshold(lambda V:V>=Vt,state='V')
    '''
    def __init__(self, thresholdfun, state=0):
        self.thresholdfun = thresholdfun # Threshold function
        self.state = state

    def __call__(self, P):
        '''
        Checks the threshold condition and returns spike times.
        P is the neuron group.
        '''
        spikes = (self.thresholdfun(P.state_(self.state)).nonzero())[0]
        #P.LS[spikes]=P.clock.t # Time of last spike (this line should be general)
        return spikes


class VariableThreshold(Threshold):
    '''
    Threshold mechanism where one state variable is compared to another.
    
    **Initialised as:** ::
    
        VariableThreshold([threshold_state=1[,state=0]])
        
    with arguments:
    
    ``threshold_state``
        The state holding the lower bound for spiking.
    ``state``
        The state that is checked.
    
    If ``x`` is the value of state variable ``threshold_state`` on neuron
    ``i`` and ``y`` is the value of state variable ``state`` on neuron
    ``i`` then neuron ``i`` will fire if ``y>x``.
    
    Typically, using this class is more time efficient than writing
    a custom thresholding operation.
    
    **Compilation**
    
    Note that if the global variable ``useweave`` is set to ``True``
    then this function will use a ``C++`` accelerated version.
    '''
    def __init__(self, threshold_state=1, state=0):
        self.threshold_state = threshold_state # State variable representing the threshold
        self.state = state
        self._useaccel = get_global_preference('useweave')
        self._cpp_compiler = get_global_preference('weavecompiler')
        self._extra_compile_args = ['-O3']
        if self._cpp_compiler == 'gcc':
            self._extra_compile_args += get_global_preference('gcc_options') # ['-march=native', '-ffast-math']

    def __call__(self, P):
        '''
        Checks the threshold condition, resets and returns spike times.
        P is the neuron group.
        '''
        if self._useaccel:
            spikes = P._spikesarray
            V = P.state_(self.state)
            Vt = P.state_(self.threshold_state)
            N = int(len(P))
            code = """
                    int numspikes=0;
                    for(int i=0;i<N;i++)
                        if(V(i)>Vt(i))
                            spikes(numspikes++) = i;
                    return_val = numspikes;
                    """
            numspikes = weave.inline(code, ['spikes', 'V', 'Vt', 'N'], \
                                     compiler=self._cpp_compiler,
                                     type_converters=weave.converters.blitz,
                                     extra_compile_args=self._extra_compile_args)
            return spikes[0:numspikes]
        else:
            return ((P.state_(self.state) > P.state_(self.threshold_state)).nonzero())[0]

    def __repr__(self):
        return 'Variable threshold mechanism'


class EmpiricalThreshold(Threshold):
    '''
    Empirical threshold, e.g. for Hodgkin-Huxley models.
    
    In empirical models such as the Hodgkin-Huxley method, after a spike
    neurons are not instantaneously reset, but reset themselves
    as part of the dynamical equations defining their behaviour. This class
    can be used to model that. It is a simple threshold mechanism that
    checks e.g. ``V>=Vt`` but it only does so for neurons that haven't
    recently fired (giving the dynamical equations time to reset
    the values naturally). It should be used in conjunction with the
    :class:`NoReset` object.
    
    **Initialised as:** ::
    
        EmpiricalThreshold([threshold=1*mV[,refractory=1*ms[,state=0[,clock]]]])

    with arguments:
    
    ``threshold``
        The lower bound for the state variable to induce a spike.
    ``refractory``
        The time to wait after a spike before checking for spikes again.
    ``state``
        The name or number of the state variable to check.
    ``clock``
        If this object is being used for a :class:`NeuronGroup` which doesn't
        use the default clock, you need to specify its clock here.
    '''
    @check_units(refractory=second)
    def __init__(self, threshold=1 * mvolt, refractory=1 * msecond, state=0, clock=None):
        self.threshold = threshold # Threshold value
        self.state = state
        clock = guess_clock(clock)
        self.refractory = int(refractory / clock.dt)
        # this assumes that if the state stays over the threshold, and say
        # refractory=5ms the user wants spiking at 0ms 5ms 10ms 15ms etc.
        if is_approx_equal(self.refractory * clock.dt, refractory) and self.refractory > 0:
            self.refractory -= 1

    def __call__(self, P):
        '''
        Checks the threshold condition, resets and returns spike times.
        P is the neuron group.
        '''
        #spikes=where((P._S[0,:]>self.Vt) & ((P.LS<P.clock.t-self.refractory) | (P.LS==P.clock.t)))[0]
        spikescond = P.state_(self.state) > self.threshold
        spikescond[P.LS[0:self.refractory]] = False
        return spikescond.nonzero()[0]
        #P.LS[spikes]=P.clock.t # Time of last spike (this line should be general)
        #return spikes

    def __repr__(self):
        return 'Empirical threshold with value=' + str(self.threshold) + " acting on state " + str(self.state)


class PoissonThreshold(Threshold):
    '''
    Poisson threshold: a spike is produced with some probability S[0]*dt,
    or S[state]*dt.
    '''
    # TODO: check the state has units in Hz
    def __init__(self, state=0):
        self.state = state

    def __call__(self, P):
        return (random.rand(len(P)) < P.state_(self.state)[:] * P.clock.dt).nonzero()[0]

    def __repr__(self):
        return 'Poisson threshold'


class HomogeneousPoissonThreshold(PoissonThreshold):
    '''
    Poisson threshold for spike trains with identical rates.
    The underlying NeuronGroup has only one state variable.
    N.B.: "homogeneous" is meant in the spatial (not temporal) sense,
    the rate may change in time.
    '''
    def __call__(self, P):
        # N.B.: is "float" necessary?
        # Other possibility to avoid sorting: use an exponential distribution
        n = random.poisson(float(len(P) * P.clock.dt * clip(P._S[self.state][0], 0, Inf))) # number of spikes
        if n > len(P):
            n = len(P)
            log_warn('brian.HomogeneousPoissonThreshold', 'HomogeneousPoissonThreshold cannot generate enough spikes.')
        spikes = sample(xrange(len(P)), n)
        spikes.sort() # necessary only for subgrouping
        return spikes



'''
This example basically replicates what the Brian PulsePacket object does,
and then compares to that object.
'''

from brian import *
from random import gauss, shuffle

# Generator for pulse packet
def pulse_packet(t, n, sigma):
    # generate a list of n times with Gaussian distribution, sort them in time, and
    # then randomly assign the neuron numbers to them
    times = [gauss(t, sigma) for i in range(n)]
    times.sort()
    neuron = range(n)
    shuffle(neuron)
    return zip(neuron, times) # returns a list of pairs (i,t)

G1 = SpikeGeneratorGroup(1000, pulse_packet(50 * ms, 1000, 5 * ms))
M1 = SpikeMonitor(G1)
PRM1 = PopulationRateMonitor(G1, bin=1 * ms)

G2 = PulsePacket(50 * ms, 1000, 5 * ms)
M2 = SpikeMonitor(G2)
PRM2 = PopulationRateMonitor(G2, bin=1 * ms)

run(100 * ms)

subplot(221)
raster_plot(M1)
subplot(223)
plot(PRM1.rate)
subplot(222)
raster_plot(M2)
subplot(224)
plot(PRM2.rate)
show()



from pylab import *
import glob, pickle, datetime

names = glob.glob('track_*.py')

def run_speedtracks(doplot=True):
    for name in names:
        mod = __import__(name.replace('.py', ''))
        print 'Running tracking from', name
        descname, results = mod.run_barrage()
        if doplot:
            figure()
            mod.plot_results(descname, results)
        dateandtime = datetime.datetime.today()
        date = datetime.date.today()
        fname = name.replace('.py', '-' + str(date) + '.pkl')
        pickle.dump((descname, results, date), open(fname, 'wb'), protocol=2)
#        date = date - datetime.timedelta(1)
#        fname = name.replace('.py', '-'+str(date)+'.pkl')
#        pickle.dump((descname, results, date), open(fname, 'wb'), protocol=2)
    if doplot:
        show()

def plot_recent(name=None):
    if name is None:
        for name in names:
            plot_recent(name)
        return
    name = name.replace('.py', '')
    mod = __import__(name)
    fnames = glob.glob(name + '-*.pkl')
    if len(fnames) == 0:
        return
    bestdate = None
    bestdescname, bestresults = None, None
    for fname in fnames:
        descname, results, date = pickle.load(open(fname, 'rb'))
        if bestdate is None or date > bestdate:
            bestdescname, bestresults, bestdate = descname, results, date
    figure()
    mod.plot_results(bestdescname, bestresults)

def plot_all(name=None):
    if name is None:
        for name in names:
            plot_all(name)
        return
    name = name.replace('.py', '')
    mod = __import__(name)
    fnames = glob.glob(name + '-*.pkl')
    for fname in fnames:
        descname, results, date = pickle.load(open(fname, 'rb'))
        figure()
        mod.plot_results(descname, results)

def meanresult(results):
    return sum(t for d, t, a in results) / len(results)

def track_mean(name=None):
    if name is None:
        for name in names:
            track_mean(name)
        legend()
        return
    name = name.replace('.py', '')
    mod = __import__(name)
    fnames = glob.glob(name + '-*.pkl')
    bestdate = None
    for fname in fnames:
        descname, results, date = pickle.load(open(fname, 'rb'))
        if bestdate is None or date > bestdate:
            bestdate = date
    means = []
    for fname in fnames:
        descname, results, date = pickle.load(open(fname, 'rb'))
        means.append(((date - bestdate).days, meanresult(results)))
    means.sort()
    meanday, meanres = zip(*means)
    plot(meanday, meanres, label=descname)


if __name__ == '__main__':
    run_speedtracks()
    #plot_recent()
    #track_mean()
    show()



"""
Analysis of spike threshold.

Loads a current clamp voltage trace, compensates (remove electrode voltage)
and analyses the spikes.
"""
from brian import *
from brian.library.electrophysiology import *
import numpy

dt=.1*ms
Vraw = numpy.load("trace.npy") # Raw current clamp trace
I = numpy.load("current.npy")
V, _ = Lp_compensate(I, Vraw, dt) # Electrode compensation

# Peaks
spike_criterion=find_spike_criterion(V)
print "Spike detected when V exceeds",float(spike_criterion/mV),"mV"
peaks=spike_peaks(V,vc=spike_criterion) # vc is optional

# Onsets (= spike threshold)
onsets=spike_onsets(V,criterion=3*dt,vc=spike_criterion) # Criterion: dV/dt>3 V/s

# Spike-triggered average of V
STA=spike_shape(V, onsets=onsets, before=100, after=100)

print "Spike duration:",float(spike_duration(V,onsets=onsets)*dt/ms),"ms"
print "Reset potential:",float(reset_potential(V,peaks=peaks)/mV),"mV"

# Spike threshold statistics
slope=slope_threshold(V,onsets=onsets,T=int(5*ms/dt))

# Subthreshold trace
subthreshold=-spike_mask(V)

t=arange(len(V))*dt
subplot(221)
plot(t/ms,V/mV,'k')
plot(t[peaks]/ms,V[peaks]/mV,".b")
plot(t[onsets]/ms,V[onsets]/mV,".r")
subplot(222)
plot(((arange(len(STA))-100)*dt)/ms,STA/mV,'k')
subplot(223)
plot(t[subthreshold]/ms,V[subthreshold]/mV,'k')
subplot(224)
plot(slope/ms,V[onsets]/mV,'.')
show()



from modelfitting import *
from brian.experimental.codegen.integration_schemes import *

#__all__ = ['modelfitting', 'print_table', 'get_spikes', 'predict',
#           'PSO', 'GA','CMAES',
#           'euler_scheme', 'rk2_scheme', 'exp_euler_scheme'
#           ]



from brian import StateUpdater, NeuronGroup, Equations, Clock, network_operation

__all__ = ['FilterbankGroup']

class FilterbankGroup(NeuronGroup):
    '''
    Allows a Filterbank object to be used as a NeuronGroup
    
    Initialised as a standard :class:`NeuronGroup` object, but with two
    additional arguments at the beginning, and no ``N`` (number of neurons)
    argument.  The number of neurons in the group will be the number of
    channels in the filterbank. (TODO: add reference to interleave/serial
    channel stuff here.)
    
    ``filterbank``
        The Filterbank object to be used by the group. In fact, any Bufferable
        object can be used.
    ``targetvar``
        The target variable to put the filterbank output into.
        
    One additional keyword is available beyond that of :class:`NeuronGroup`:
    
    ``buffersize=32``
        The size of the buffered segments to fetch each time. The efficiency
        depends on this in an unpredictable way, larger values mean more time
        spent in optimised code, but are worse for the cache. In many cases,
        the default value is a good tradeoff. Values can be given as a number
        of samples, or a length of time in seconds.
        
    Note that if you specify your own :class:`Clock`, it should have
    1/dt=samplerate.
    '''
    
    def __init__(self, filterbank, targetvar, *args, **kwds):
        self.targetvar = targetvar
        self.filterbank = filterbank
        filterbank.buffer_init()

        # update level keyword
        kwds['level'] = kwds.get('level', 0)+1
    
        # Sanitize the clock - does it have the right dt value?
        if 'clock' in kwds:
            if int(1/kwds['clock'].dt)!=int(filterbank.samplerate):
                raise ValueError('Clock should have 1/dt=samplerate')
        else:
            kwds['clock'] = Clock(dt=1/filterbank.samplerate)        
        
        buffersize = kwds.pop('buffersize', 32)
        if not isinstance(buffersize, int):
            buffersize = int(buffersize*self.samplerate)
        self.buffersize = buffersize
        self.buffer_pointer = buffersize
        self.buffer_start = -buffersize
        
        NeuronGroup.__init__(self, filterbank.nchannels, *args, **kwds)
        
        @network_operation(when='start', clock=self.clock)
        def apply_filterbank_output():
            if self.buffer_pointer>=self.buffersize:
                self.buffer_pointer = 0
                self.buffer_start += self.buffersize
                self.buffer = self.filterbank.buffer_fetch(self.buffer_start, self.buffer_start+self.buffersize)
            setattr(self, targetvar, self.buffer[self.buffer_pointer, :])
            self.buffer_pointer += 1
        
        self.contained_objects.append(apply_filterbank_output)
        
    def reinit(self):
        NeuronGroup.reinit(self)
        self.filterbank.buffer_init()
        self.buffer_pointer = self.buffersize
        self.buffer_start = -self.buffersize



'''
Network to demonstrate stopping a simulation during a run

Have a fully connected network of integrate and fire neurons
with input fed by a group of Poisson neurons with a steadily
increasing rate, want to determine the point in time at which
the network of integrate and fire neurons switches from no
firing to all neurons firing, so we have a network_operation
called stop_condition that calls the stop() function if the
monitored network firing rate is above a minimum threshold.
'''

from brian import *

clk = Clock()

Vr = 0 * mV
El = 0 * mV
Vt = 10 * mV
tau = 10 * ms
weight = 0.2 * mV
duration = 100 * msecond
max_input_rate = 10000 * Hz
num_input_neurons = 1000
input_connection_p = 0.1
rate_per_neuron = max_input_rate / (num_input_neurons * input_connection_p)

P = PoissonGroup(num_input_neurons, lambda t: rate_per_neuron * (t / duration))

G = NeuronGroup(1000, model='dV/dt=-(V-El)/tau : volt', threshold=Vt, reset=Vr)
G.V = Vr + (Vt - Vr) * rand(len(G))

CPG = Connection(P, G, weight=weight, sparseness=input_connection_p)

CGG = Connection(G, G, weight=weight)

MP = PopulationRateMonitor(G, bin=1 * ms)

@network_operation
def stop_condition():
    if MP.rate[-1] * Hz > 10 * Hz:
        stop()

run(duration)

print "Reached population rate>10 Hz by time", clk.t, "+/- 1 ms."



import pycuda.autoinit as autoinit
import pycuda.driver as drv
from pycuda.gpuarray import GPUArray
from pycuda import gpuarray
import numpy, pylab, time

N = 512 * 8
spikes = [i for i in range(N) if numpy.random.rand() < 0.1]
blocksize = 512
block = (blocksize, 1, 1)
grid = (N / blocksize, 1)
repeats = 1000

print 'N:', N
print 'numspikes:', len(spikes), 'proportion', float(len(spikes)) / N
print 'repeats:', repeats
print 'block:', block, 'grid:', grid

mod = drv.SourceModule("""
__global__ void propagate(int *spikes, int numspikes, float *v, float *W, int N)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    for(int j=0; j<numspikes; j++)
        v[i] += W[i+N*spikes[j]];
}
""")

propagate = mod.get_function("propagate")

W = numpy.random.randn(N, N)
W_gpu = gpuarray.to_gpu(numpy.array(W, dtype=numpy.float32))

v = gpuarray.to_gpu(numpy.zeros(N, dtype=numpy.float32))
v_pre = v.get()

gpu_spikes = drv.mem_alloc(4 * len(spikes))
spikes = numpy.array(spikes, dtype=int)
drv.memcpy_htod(gpu_spikes, spikes)

start = time.time()
for _ in xrange(repeats):
    propagate(gpu_spikes, numpy.int32(len(spikes)), v, W_gpu, numpy.int32(N),
              block=block, grid=grid)
autoinit.context.synchronize()
print 'GPU propagation:', time.time() - start
v_post_gpu = v.get()
gputime = time.time() - start
print 'GPU propagation and copy:', gputime

start = time.time()
v_post = v_pre
for _ in xrange(repeats):
    for i in spikes:
        v_post += W[i, :]
cputime = time.time() - start
print 'CPU propagation:', cputime, '(' + str(int(cputime / gputime)) + 'x)'

print 'Max abs difference CPU-GPU:', numpy.amax(numpy.abs(v_post_gpu - v_post))



'''
Polychronization: Computation with Spikes
Eugene M. Izhikevich
Neural Computation, 2006

See Izhikevich_2006_Polychronization.py for an explanation of this code.

In this version, we use various tricks to make the simulation much faster,
including inline C++ code. We also save the progress as the simulation runs,
and analyse it at the end for polychronous groups.

Note that you will need to create two folders data and imgs to run this
script.
'''
import brian_no_units_no_warnings
from brian import *
import random
from scipy import weave
import cPickle as pickle

try:
    imported_data = pickle.load(open('data/izhikevich.pickle', 'rb'))
    defaultclock.t = imported_data['t']
    print 'Starting from saved progress at time', defaultclock.t
except IOError:
    imported_data = {}

dt = defaultclock.dt = 0.1 * ms

M = 100                # number of synapses per neuron
D = 20 * ms              # maximal conduction delay
D_min = 1 * ms           # minimal conduction delay
D_i = 1 * ms             # inhibitory neuron conduction delay

# excitatory neurons
Ne = 800
a_e = 0.02 / ms
d_e = 8 * mV / ms
s_e = 6 * mV / dt # ms
# inhibitory neurons
Ni = 200
a_i = 0.1 / ms
d_i = 2 * mV / ms
s_i = -5 * mV / dt # ms
# all neurons
b = 0.2 / ms
c = -65 * mV
sm = 10 * mV / dt # ms             # maximal synaptic strength
N = Ne + Ni

thalamic_input_weight = 20 * mV / dt # ms

eqs = Equations('''
dv/dt = (0.04/ms/mV)*v**2+(5/ms)*v+140*mV/ms-u+I : volt
du/dt = a*(b*v-u)                                : volt/second
a                                                : 1/second
d                                                : volt/second
I                                                : volt/second
''')
reset = '''
v = c
u += d
'''
threshold = 30 * mV

G = NeuronGroup(N, eqs, threshold=threshold, reset=reset)

def make_Izhikevich_scheme():
    c1 = 0.04 / ms / mV
    c2 = 5 / ms
    c3 = 140 * mV / ms
    # Izhikevich's numerical integration scheme
    def Izhikevich_scheme(G):
        G.v += (0.5 * dt) * (c1 * G.v ** 2 + c2 * G.v + c3 - G.u + G.I)
        G.v += (0.5 * dt) * (c1 * G.v ** 2 + c2 * G.v + c3 - G.u + G.I)
        G.u += dt * G.a * (b * G.v - G.u)
    weave_code = '''
    for(int i=0; i<N; i++){
        v[i] += (0.5*dt)*(c1*v[i]*v[i]+c2*v[i]+c3-u[i]+I[i]);
        v[i] += (0.5*dt)*(c1*v[i]*v[i]+c2*v[i]+c3-u[i]+I[i]);
        u[i] += dt*a[i]*(b*v[i]-u[i]); 
    }
    '''
    weave_vars = (G.a, G.v, G.u, G.I, c1, c2, c3)
    weave_compiler = get_global_preference('weavecompiler')
    extra_compile_args = ['-O3']
    #if weave_compiler == 'gcc':
    #    extra_compile_args += get_global_preference('gcc_options') # ['-march=native', '-ffast-math']
    #    extra_compile_args.remove('-march=native')
    def weave_Izhikevich_scheme(G):
        a, v, u, I, c1, c2, c3 = weave_vars
        weave.inline(weave_code, ['a', 'v', 'u', 'I', 'c1', 'c2', 'c3', 'b', 'dt', 'N'],
                     compiler=weave_compiler, extra_compile_args=extra_compile_args)
    if get_global_preference('useweave'):
        return weave_Izhikevich_scheme
    return Izhikevich_scheme
G._state_updater = make_Izhikevich_scheme()

Ge = G.subgroup(Ne)
Gi = G.subgroup(Ni)

G.v = c
G.u = b * c
Ge.a = a_e
Ge.d = d_e
Gi.a = a_i
Gi.d = d_i

zeromVms = 0 * mV / ms
change_every = int(1*ms/dt)
change_time = 0
thal_ind = 0
@network_operation(when='before_connections')
def thalamic_input():
    global change_time, thal_ind
    G.I = zeromVms
    #G.I[thal_ind] = float(thalamic_input_weight)
    if change_time==0:
        change_time = change_every
        thal_ind = randint(N)
        G.I[thal_ind] = float(thalamic_input_weight)
    change_time -= 1

if imported_data:
    G._S[:] = imported_data['G._S']

Ce = Connection(Ge, G, 'I', delay=True, max_delay=D)
Ci = Connection(Gi, Ge, 'I', delay=D_i)
Ce_deriv = Connection(Ge, G, 'I', delay=True, max_delay=D)
forget(Ce_deriv)

if imported_data:
    exc_inds = imported_data['exc_inds']
    exc_weights = imported_data['exc_weights']
    exc_delays = imported_data['exc_delays']
    inh_inds = imported_data['inh_inds']
    inh_weights = imported_data['inh_weights']
    for i in xrange(Ne):
        Ce[i, exc_inds[i]] = exc_weights[i]
        Ce.delay[i, exc_inds[i]] = exc_delays[i]
        Ce_deriv[i, exc_inds[i]] = exc_weights[i]
        Ce_deriv.delay[i, exc_inds[i]] = exc_delays[i]
    for i in xrange(Ni):
        Ci[i, inh_inds[i]] = inh_weights[i]
else:
    for i in xrange(Ne):
        inds = random.sample(xrange(N), M)
        Ce[i, inds] = s_e * ones(len(inds))
        Ce.delay[i, inds] = rand(len(inds)) * (D - D_min) + D_min
    for i in xrange(Ni):
        inds = random.sample(xrange(Ne), M)
        Ci[i, inds] = s_i * ones(len(inds))
    for i in xrange(Ne):
        Ce_deriv[i, :] = Ce[i, :]
        Ce_deriv.delay[i, :] = Ce.delay[i, :]

artificial_wmax = 1e10 * mV / ms
stdp = ExponentialSTDP(Ce_deriv,
                       taup=20 * ms, taum=20 * ms,
                       Ap=(0.1 * mV / ms / artificial_wmax),
                       Am=(-0.12 * mV / ms / artificial_wmax),
                       wmin= -artificial_wmax,
                       wmax=artificial_wmax,
                       interactions='nearest'
                       )

Ce_deriv.compress()
Ce_deriv.W.alldata[:] = 0
weight_derivative_const = 0.01 * mV / ms
@network_operation(clock=EventClock(dt=1 * second))
def update_weights_from_derivative():
    Ce.W.alldata[:] = clip(Ce.W.alldata + weight_derivative_const + Ce_deriv.W.alldata, 0, sm)
    Ce_deriv.W.alldata[:] *= 0.9

M = SpikeMonitor(G)

def autocorrelogram(T, width=40 * ms, bin=1 * ms):
    b = bincount(array(array(T) / bin, dtype=int))
    ac = correlate(b, b, mode='full')
    ac = ac[len(ac) / 2 - int(width / bin):len(ac) / 2 + int(width / bin) + 1]
    edges = arange(-width, width + bin, bin)
    return ac, edges

@network_operation(clock=EventClock(dt=10 * second, t=defaultclock.t + 10 * second))
def plot_recent_output():
    if not M.spikes:
        return
    clf()
    subplot(211)
    raster_plot(M, ms=1)
    s = str(int((defaultclock.t + .5 * ms) / second))
    title('t = ' + s + ' s')
    subplot(212)
    ac, edges = autocorrelogram([t for i, t in M.spikes], width=.3 * second, bin=5 * ms)
    emin = min((edges > 1. / 100).nonzero()[0])
    plot(1 / edges[emin:], ac[emin:])
    gcf().savefig('imgs/iz' + s + '.png')
    pickle.dump(M.spikes, open('data/iz_spikes' + s + '.pickle', 'wb'), -1)
    M.reinit()
    print 'Plotted up to time', s

@network_operation(clock=EventClock(dt=100 * second, t=defaultclock.t + 100 * second))
def save_progress():
    s = str(int((defaultclock.t + .5 * ms) / second))
    imported_data['G._S'] = G._S
    imported_data['t'] = defaultclock.t
    exc_inds = [Ce.W[i, :].ind for i in range(Ne)]
    exc_weights = [asarray(Ce.W[i, :]) for i in range(Ne)]
    exc_delays = [asarray(Ce.delay[i, :]) for i in range(Ne)]
    inh_inds = [Ci.W[i, :].ind for i in range(Ni)]
    inh_weights = [asarray(Ci.W[i, :]) for i in range(Ni)]
    imported_data['exc_inds'] = exc_inds
    imported_data['exc_weights'] = exc_weights
    imported_data['exc_delays'] = exc_delays
    imported_data['inh_inds'] = inh_inds
    imported_data['inh_weights'] = inh_weights
    pickle.dump(imported_data, open('data/izhikevich.pickle', 'wb'), -1)
    print 'Saved progress up to time', s

run(3600 * second - defaultclock.t, report='stderr')




'''
Example with short term plasticity model
Neurons with regular inputs and depressing synapses
'''
from brian import *

tau_e = 3 * ms
taum = 10 * ms
A_SE = 250 * pA
Rm = 100 * Mohm
N = 10

eqs = '''
dx/dt=rate : 1
rate : Hz
'''

input = NeuronGroup(N, model=eqs, threshold=1., reset=0)
input.rate = linspace(5 * Hz, 30 * Hz, N)

eqs_neuron = '''
dv/dt=(Rm*i-v)/taum:volt
di/dt=-i/tau_e:amp
'''
neuron = NeuronGroup(N, model=eqs_neuron)

C = Connection(input, neuron, 'i')
C.connect_one_to_one(weight=A_SE)
stp = STP(C, taud=1 * ms, tauf=100 * ms, U=.1) # facilitation
#stp=STP(C,taud=100*ms,tauf=10*ms,U=.6) # depression
trace = StateMonitor(neuron, 'v', record=[0, N - 1])

run(1000 * ms)
subplot(211)
plot(trace.times / ms, trace[0] / mV)
title('Vm')
subplot(212)
plot(trace.times / ms, trace[N - 1] / mV)
title('Vm')
show()



import cloud
import time
from brian import *
from multimap import *

def stdp_example(N=1000):
    taum = 10 * ms
    tau_pre = 20 * ms
    tau_post = tau_pre
    Ee = 0 * mV
    vt = -54 * mV
    vr = -60 * mV
    El = -74 * mV
    taue = 5 * ms
    F = 15 * Hz
    gmax = .01
    dA_pre = .01
    dA_post = -dA_pre * tau_pre / tau_post * 1.05

    eqs_neurons = '''
    dv/dt=(ge*(Ee-vr)+El-v)/taum : volt   # the synaptic current is linearized
    dge/dt=-ge/taue : 1
    '''

    input = PoissonGroup(N, rates=F)
    neurons = NeuronGroup(1, model=eqs_neurons, threshold=vt, reset=vr)
    synapses = Connection(input, neurons, 'ge', weight=rand(len(input), len(neurons)) * gmax)
    neurons.v = vr

    #stdp=ExponentialSTDP(synapses,tau_pre,tau_post,dA_pre,dA_post,wmax=gmax)
    ## Explicit STDP rule
    eqs_stdp = '''
    dA_pre/dt=-A_pre/tau_pre : 1
    dA_post/dt=-A_post/tau_post : 1
    '''
    dA_post *= gmax
    dA_pre *= gmax
    stdp = STDP(synapses, eqs=eqs_stdp, pre='A_pre+=dA_pre;w+=A_post',
              post='A_post+=dA_post;w+=A_pre', wmax=gmax)

    rate = PopulationRateMonitor(neurons)

    run(30 * second, report='text')

    return mean(rate.rate)

ilist = [1000 for _ in xrange(max_parallel)]
jids = multimap(stdp_example, ilist)

t1 = time.clock()
results = retrieve(jids)
t = time.clock() - t1

print "All jobs done in %.1f seconds, results :" % t
print results




import os
os.chdir('../../../docs_sphinx') # work from docs_sphinx/ directory until we make the big change
# if this doesn't work, run build_html_clean.py instead
os.system('sphinx-build -b html . ../docs')



# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
######### PHYSICAL UNIT CLASSES ###################
#------------------------------------ Dan Goodman -
#todo: More additional derived units (for the lookup table)

"""Defines physical units and quantities

The standard way to use this class is as follows:

V = 3 * volt
I = 2 * amp
R=V/I
print R

will return

1.5 ohm

The following fundamental units are defined:

metre, kilogram, second, amp, kelvin, mole, candle

And these additional basic units:

radian, steradian, hertz, newton, pascal, joule, watt,
coulomb, volt, farad, ohm, siemens, weber, tesla, henry,
celsius, lumen, lux, becquerel, gray, sievert, katal,
gram, gramme

Additionally, it includes all scaled versions of these
units using the standard SI prefixes (see the documentation
for the Unit class for more details), e.g. uamp,
mmetre, etc. It also includes the second and third powers
of each of these units, e.g. mvolt2 = mvolt*mvolt,
metre3 = metre**3, etc.

The module also defines these classes:

-- Dimension
        Stores the physical dimensions (length, mass, etc.)
-- DimensionMismatchError
        Exception raised if you try to add inconsistent units,
        etc.
-- Quantity
        The class of a value with a unit
-- Unit
        The class of the defined units like mvolt, etc.
-- UnitRegistry
        Stores 'known' units for printing

These functions:

-- get_dimensions(x)
        Returns the dimensions of a quantity or number x
-- have_same_dimensions(x,y)
        Tests if x and y have the same dimensions
-- is_dimensionless(x)
        Tests if x is dimensionless
-- display_in_unit(x,u)
        Displays quantity x in unit u
-- register_new_unit(u)
        Add a new unit u to the list of 'known' units for
        printing purposes
-- get_unit(x)
        Returns the fundamental unit of value x if one is known, or
        simply the value 1 with dimensions of x if none is known

And this decorator for function argument checking:

-- check_units(...)

If you want to use shorter named units, import the stdunits
module, which defines things like mV for mvolt, etc. They
are not included by default in the units module because of
the potential for variable name clashes.
"""

__origall__ = ['Dimension', 'Scale', 'DimensionMismatchError',
           'get_dimensions', 'is_dimensionless', 'have_same_dimensions',
           'display_in_unit', 'Quantity', 'Unit', 'register_new_unit',
           'check_units', 'is_scalar_type', 'get_unit', 'get_unit_fast',
           'scalar_representation']

__all__ = __origall__ + []

from brian_unit_prefs import bup
from operator import isNumberType, isSequenceType
from itertools import izip
import math, numpy
from utils.approximatecomparisons import *
import types
from functools import *
import sys
# Note that the decorator module below is used to provide signature preserving
# decorators, but it has the unfortunate side effect of messing up the tracebacks
# because it uses eval, so we only use it when we want to generate documentation,
# i.e. if 'sphinx' or 'docutils' or 'epydoc' are loaded.
# TODO: this has stopped working anyway, so it's now removed
#try:
#    import decorator
#    use_decorator = 'sphinx' in sys.modules or 'docutils' in sys.modules or 'epydoc' in sys.modules
#except:
#    use_decorator = False

# SI dimensions (see table at end of file) and various descriptions,
# each description maps to an index i, and the power of each dimension
# is stored in the variable dims[i]
_di = { "Length":0, "length": 0, "metre":0, "metres":0, "metre": 0, "metres":0, "metre":0, "metres":0, "metre": 0, "metres":0, "m": 0, \
       "Mass":1, "mass": 1, "kilogram":1, "kilograms":1, "kilogram": 1, "kilograms":1, "kg": 1, \
       "Time":2, "time": 2, "second":2, "seconds":2, "second": 2, "seconds":2, "s": 2, \
       "Electric Current":3, "Electric Current":3, "electric current": 3, "Current":3, "current":3, "ampere":3, "amperes":3, "ampere": 3, "amperes":3, "A": 3, \
       "Temperature":4, "temperature": 4, "kelvin":4, "kelvins":4, "kelvin": 4, "kelvins":4, "K": 4, \
       "Quantity of Substance":5, "Quantity of substance": 5, "quantity of substance": 5, "Substance":5, "substance":5, "mole":5, "moles":5, "mole": 5, "moles":5, "mol": 5, \
       "Luminosity":6, "luminosity": 6, "candle":6, "candles":6, "candle": 6, "candles":6, "cd": 6 }
_ilabel = ["m", "kg", "s", "A", "K", "mol", "cd"]
# SI unit _prefixes, see table at end of file
_siprefixes = {"y":1e-24, "z":1e-21, "a":1e-18, "f":1e-15, "p":1e-12, "n":1e-9, "u":1e-6, "m":1e-3, "c":1e-2, "d":1e-1, \
            "":1, \
            "da":1e1, "h":1e2, "k":1e3, "M":1e6, "G":1e9, "T":1e12, "P":1e15, "E":1e18, "Z":1e21, "Y":1e24}


class Dimension(object):
    '''Stores the indices of the 7 basic SI unit dimension (length, mass, etc.)
    
    Provides a subset of arithmetic operations appropriate to dimensions:
    multiplication, division and powers, and equality testing.
    
    Methods:
    
    is_dimensionless() returns Boolean value
    
    Notes:
    
    Most users shouldn't use this class directly, but instead write things
    like:
    
    x = 3 * mvolt, etc.
    '''
    __slots__ = ["_dims"]
    #### INITIALISATION ####
    def __init__(self, *args, **keywords):
        """Initialise Dimension object with a vector or keywords
        
        Call as Dimension(list), Dimension(keywords) or Dimension(dim)
        
        list -- a list with the indices of the 7 elements of an SI dimension
        keywords -- a sequence of keyword=value pairs where the keywords are
          the names of the SI dimensions, or the standard unit
        dim -- a dimension object to copy
        
        Examples:
        
        The following are all definitions of the dimensions of force
        
        Dimension(length=1, mass=1, time=-2)
        Dimension(m=1, kg=1, s=-2)
        Dimension([1,1,-2,0,0,0,0])
        
        The 7 units are (in order):
        
        Length, Mass, Time, Electric Current, Temperature,
        Quantity of Substance, Luminosity
        
        and can be referred to either by these names or their SI unit names,
        e.g. length, metre, and m all refer to the same thing here.
        """
        if len(args):
            if isSequenceType(args[0]) and len(args[0]) == 7:
                # initialisation by list
                self._dims = args[0]
            elif isinstance(args[0], Dimension):
                # initialisation by another dimension object
                self._dims = args[0]._dims
        else:
            # initialisation by keywords
            self._dims = [0, 0, 0, 0, 0, 0, 0]
            for k in keywords.keys():
                # _di stores the index of the dimension with name 'k'
                self._dims[_di[k]] = keywords[k]
    #### METHODS ####
    def get_dimension(self, d):
        """Returns the list of dimension indices.
        
        See documentation for __init__.
        """
        return self._dims[_di[d]]

    def set_dimension(self, d, value):
        """Sets the list of dimension indices.
        
        See documentation for __init__.
        """
        self._dims[_di[d]] = value

    def is_dimensionless(self):
        """Tells you whether the object is dimensionless."""
        return sum([x == 0 for x in self._dims]) == 7
    #### REPRESENTATION ####
    def __repr__(self):
        return self.__str__()

    def __str__(self):
        """String representation in basic SI units, or 1 for dimensionless."""
        s = ""
        for i in range(len(self._dims)):
            if self._dims[i]:
                s += _ilabel[i]
                if self._dims[i] != 1: s += "^" + str(self._dims[i])
                s += " "
        if not len(s): return "1"
        return s.strip()
    #### ARITHMETIC ####
    # Note that none of the dimension arithmetic objects do sanity checking
    # on their inputs, although most will throw an exception if you pass the
    # wrong sort of input
    def __mul__(self, value):
        return Dimension([x + y for x, y in izip(self._dims, value._dims)])

    def __div__(self, value):
        return Dimension([x - y for x, y in izip(self._dims, value._dims)])

    def __truediv__(self, value):
        return self.__div__(value)

    def __pow__(self, value):
        return Dimension([x * value for x in self._dims])

    def __imul__(self, value):
        self._dims = [x + y for x, y in izip(self._dims, value._dims)]
        return self

    def __idiv__(self, value):
        self._dims = [x - y for x, y in izip(self._dims, value._dims)]
        return self

    def __itruediv__(self, value):
        return self.__idiv__(value)

    def __ipow__(self, value):
        self._dims = [x * value for x in self._dims]
        return self
    #### COMPARISON ####
    def __eq__(self, value):
        #return sum([x==y for x,y in izip(self._dims,value._dims)])==7
        return sum([is_within_absolute_tolerance(x, y) for x, y in izip(self._dims, value._dims)]) == 7

    def __ne__(self, value):
        return not self.__eq__(value)
    #### MAKE DIMENSION PICKABLE ####
    def __getstate__(self):
        return self._dims

    def __setstate__(self, state):
        self._dims = state


class Scale(object):
    """Stores the scale factor for each SI dimension.
    
    Probably would only be very rarely used by a user, but might
    conceivably be useful in certain circumstances.
    
    Methods:
    
    -- Initialisation by list of keywords
    -- scale_factor(dim) gives the overall scaling for a value in
       dimension dim
    -- unit_representation(dim) gives a string representation of
       the unit defined by the Scale object applied to dimension
       dim 
    """
    __slots__ = ["scale"]

    def __init__(self, *args, **keywords):
        """Initialise by list of scales or keywords, see Dimension documentation
        
        e.g. Scale(length="m", time="u") =
             Scale(["m","","u","","","",""])
        corresponds to measuring the unit of length at the milli scale
        and the unit of time at the u scale.
        """
        self.scale = [ "", "", "", "", "", "", "" ]
        for k in keywords:
            self.scale[_di[k]] = keywords[k]

    def scale_factor(self, dim):
        """Returns the scaling factor for dimension dim
        
        For example, if the scale factor of length is milli, and the
        dimensions of dim are length^2 then the scale factor will be
        0.001^2.
        """
        sf = 1
        for s, i in izip(self.scale, dim._dims):
            if i: sf *= _siprefixes[s] ** i
        return sf

    def unit_representation(self, dim):
        """Returns a representation of the dimension dim at this scale
        
        For example, if the scale factor of length is milli, and the
        dimensions of dim are length^2 then this will return mm^2.
        """
        s = ""
        for i in range(7):
            if dim._dims[i]:
                s += self.scale[i] + _ilabel[i]
                if dim._dims[i] != 1:
                    s += "^" + str(dim._dims[i])
                s += " "
        return s.strip()


class DimensionMismatchError(Exception):
    """Exception class for attempted operations with inconsistent dimensions
    
    For example, ``3*mvolt + 2*amp`` raises this exception. The purpose of this
    class is to help catch errors based on incorrect units. The exception will
    print a representation of the dimensions of the two inconsistent objects
    that were operated on. If you want to check for inconsistent units in your
    code, do something like::
    
        try:
            ...
            your code here
            ...
        except DimensionMismatchError, inst:
            ...
            cleanup code here, e.g.
            print "Found dimension mismatch, details:", inst
            ...
    """
    def __init__(self, description, *dims):
        """Raise as DimensionMismatchError(desc,dim1,dim2,...)
        
        desc -- a description of the type of operation being performed, e.g.
                Addition, Multiplication, etc.
        dim -- the dimensions of the objects involved in the operation, any
               number of them is possible
        """
        self._dims = dims
        self.desc = description

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        s = self.desc + ", dimensions were "
        for d in self._dims:
            s += "(" + str(d) + ") "
        return s

def is_scalar_type(obj):
    """Tells you if the object is a 1d number type
    
    This function is mostly used internally by the module for
    argument type checking. A scalar type can be considered
    a dimensionless quantity (see the documentation for
    Quantity for more information).
    """
    return isNumberType(obj) and not isSequenceType(obj)

def get_dimensions(obj):
    """Returns the dimensions of any object that has them.
    
    Slightly more general than obj.get_dimensions() because it will return
    a new dimensionless Dimension() object if the object is of number type
    but not a Quantity (e.g. a float or int).
    """
    if isNumberType(obj) and not isinstance(obj, Quantity): return Dimension()
    return obj.get_dimensions()


def is_dimensionless(obj):
    """Tests if a scalar value is dimensionless or not, returns a ``bool``.
    
    Note that the syntax may change in later releases of Brian, with tighter
    integration of scalar and array valued quantities.
    """
    return get_dimensions(obj) == Dimension()


def have_same_dimensions(obj1, obj2):
    """Tests if two scalar values have the same dimensions, returns a ``bool``.
    
    Note that the syntax may change in later releases of Brian, with tighter
    integration of scalar and array valued quantities.
    """
    return get_dimensions(obj1) == get_dimensions(obj2)

def display_in_unit(x, u):
    """String representation of the object x in unit u.
    """
    if not have_same_dimensions(x, u):
        raise DimensionMismatchError("Non-matching unit for function display_in_unit", get_dimensions(x), get_dimensions(u))
    s = str(float(x / u)) + " "
    if not is_dimensionless(u):
        if isinstance(u, Unit):
            s += str(u)
        else:
            s += str(u.dim)
    return s.strip()

def quantity_with_dimensions(floatval, dims):
    return Quantity.with_dimensions(floatval, dims)


class Quantity(numpy.float64):
    """A number with an associated physical dimension.
    
    In most cases, it is not necessary to create a :class:`Quantity` object
    by hand, instead use the constant unit names ``second``, ``kilogram``,
    etc. The details of how :class:`Quantity` objects work is subject to
    change in future releases of Brian, as we plan to reimplement it
    in a more efficient manner, more tightly integrated with numpy. The
    following can be safely used:
    
    * :class:`Quantity`, this name will not change, and the usage
      ``isinstance(x,Quantity)`` should be safe.
    * The standard unit objects, ``second``, ``kilogram``, etc.
      documented in the main documentation will not be subject
      to change (as they are based on SI standardisation).
    * Scalar arithmetic will work with future implementations.
    """

    # This documentation is subject to change.
    """
    This is the main user class for the units module, although
    in most cases it is not necessary to initialise a new
    quantity by hand (see construction below for details).
    
    The Quantity class defines arithmetic operations which
    check for consistency of dimensions and raise the
    DimensionMismatchError exception if they are inconsistent.
    
    The class also defines default and other representations
    of a number for printing purposes.
    
    Typical usage:
    
    I = 3 * amp # I is a Quantity object
    R = 2 * ohm # same for R
    print I*R # displays "6 V"
    print (I*R).in_unit(mvolt) # displays "6000 mV"
    print (I*R)/mvolt # displays "6000"
    x = I + R # raises DimensionMismatchError
    
    See the documentation on the Unit class for more details
    about the available unit names like mvolt, etc.
    
    Casting rules:
    
    The three rules that define the casting operations for
    Quantity object are:
    
    1. Quantity op Quantity = Quantity
        - Performs dimension checking if appropriate
    2. Scalar op Quantity = Quantity 
        - Assumes that the scalar is dimensionless
    3. other op Quantity = other
        - The Quantity object is downcast to a float
    
    Scalar types are 1 dimensional number types, including float, int, etc.
    but not array.
    
    The Quantity class is a derived class of float, so many other operations
    will also downcast to float. For example, sin(x) where x is a quantity
    will return sin(float(x)) without doing any dimension checking.

    Construction details:
    
    x = Quantity(value) returns a dimensionless object, you can then
        set the dimensions via x.set_dimensions(dim)
        
    x = Quantity.with_dimensions(value,dim) returns an object with
        floating point value value, and dimensions dim, see the
        documentation for Quantity.with_dimensions(...) for more.

    Static constructors:
    
    -- with_dimensions(dim)
    -- with_dimensions(keywords...)
    
    Methods:
    
    -- get_dimensions() return Dimension
    -- set_dimensions(dim)
    -- is_dimensionless() return boolean
    -- at_scale(scale) return string
    -- has_same_dimensions(other) return boolean
    -- in_unit(unit) return string
    -- in_best_unit() return string
    """
    __slots__ = ["dim"]
    #### CONSTRUCTION ####
    def __init__(self, value):
        """Initialises as dimensionless
        """
        super(Quantity, self).__init__()
        self.dim = Dimension()
    @staticmethod
    def with_dimensions(value, *args, **keywords):
        """Static method to create a Quantity object with dimensions
        
        Use as Quantity.with_dimensions(value,dim),
               Quantity.with_dimensions(value,dimlist) or
               Quantity.with_dimensions(value,keywords...)
               
        -- value is a float or other scalar type
        -- dim is a dimension object
        -- dimlist, keywords (see the Dimension constructor)
        
        e.g.
        
        x = Quantity.with_dimensions(2,Dimension(length=1))
        x = Quantity.with_dimensions(2,length=1)
        x = 2 * metre
        
        all define the same object.
        """
        x = Quantity(value)
        if len(args) and isinstance(args[0], Dimension):
            x.set_dimensions(args[0])
        else:
            x.set_dimensions(Dimension(*args, **keywords))
        return x
    #### METHODS ####
    def get_dimensions(self):
        """Returns the dimensions of this object
        """
        return self.dim

    def set_dimensions(self, dim):
        """Set the dimensions of this object
        """
        self.dim = dim

    def is_dimensionless(self):
        """Tells you whether this is a dimensionless object
        """
        return self.dim.is_dimensionless()

    def at_scale(self, scale):
        """Returns a string representation at given scale
        """
        return str(float(self) / scale.scale_factor(self.dim)) + " " + scale.unit_representation(self.dim)

    def has_same_dimensions(self, other):
        """Tells you if this object has the same dimensions as another.
        """
        return self.dim == get_dimensions(other)

    def in_unit(self, u):
        """String representation of the object in unit u.
        """
        if not self.has_same_dimensions(u):
            raise DimensionMismatchError("Non-matching unit for method in_unit", self.dim, u.dim)
        s = str(float(self / u)) + " "
        if not u.is_dimensionless():
            if isinstance(u, Unit):
                s += str(u)
            else:
                s += str(u.dim)
        return s.strip()

    def in_best_unit(self, *regs):
        """String representation of the object in the 'best unit'
        
        For more information, see the documentation for the UnitRegistry
        class. Essentially, this looks at the value of the quantity for
        all 'known' matching units (e.g. mvolt, namp, etc.) and returns
        the one with the most compact representation. Standard units are
        built in, but you can register new units for consideration. 
        """
        u = _get_best_unit(self, *regs)
        return self.in_unit(u)
    #### METHODS (NUMERICAL) ####
    def sqrt(self):
        return self ** 0.5

    def log(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.log(float(self)), self.dim)
        raise DimensionMismatchError('log', self.dim)

    def exp(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.exp(float(self)), self.dim)
        raise DimensionMismatchError('exp', self.dim)

    def sin(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.sin(float(self)), self.dim)
        raise DimensionMismatchError('sin', self.dim)

    def cos(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.cos(float(self)), self.dim)
        raise DimensionMismatchError('cos', self.dim)

    def tan(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.tan(float(self)), self.dim)
        raise DimensionMismatchError('tan', self.dim)

    def asin(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.asin(float(self)), self.dim)
        raise DimensionMismatchError('asin', self.dim)

    def acos(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.acos(float(self)), self.dim)
        raise DimensionMismatchError('acos', self.dim)

    def atan(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.atan(float(self)), self.dim)
        raise DimensionMismatchError('atan', self.dim)
    arcsin = asin
    arccos = cos
    arctan = tan

    def sinh(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.sinh(float(self)), self.dim)
        raise DimensionMismatchError('sinh', self.dim)

    def cosh(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.cosh(float(self)), self.dim)
        raise DimensionMismatchError('cosh', self.dim)

    def tanh(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(math.tanh(float(self)), self.dim)
        raise DimensionMismatchError('tanh', self.dim)

    def arcsinh(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(numpy.arcsinh(float(self)), self.dim)
        raise DimensionMismatchError('sinh', self.dim)

    def arccosh(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(numpy.arccosh(float(self)), self.dim)
        raise DimensionMismatchError('cosh', self.dim)

    def arctanh(self):
        if self.is_dimensionless():
            return Quantity.with_dimensions(numpy.arctanh(float(self)), self.dim)
        raise DimensionMismatchError('tanh', self.dim)
    #### REPRESENTATION ####
    def __repr__(self):
        #return self.in_best_unit()
        return self.__str__()

    def __str__(self):
        #s = super(Quantity,self).__str__()
        #if not self.is_dimensionless(): s += " " + str(self.dim)
        #return s
        return self.in_best_unit()
        #return str(float(self))+'*'+str(get_unit(self))
    #### ARITHMETIC ####
    # Arithmetic operations implement the following set of rules for
    # determining casting:
    # 1. Quantity op Quantity returns Quantity (and performs dimension checking if appropriate)
    # 2. Scalar op Quantity returns Quantity (and performs dimension checking assuming Scalar is dimensionless)
    # 3. other op Quantity returns other (Quantity is downcast to float)
    # Scalar types are those for which is_scalar_type() returns True, including float, int, long, complex but not array
    def __mul__(self, other):
        # This code, like all the other arithmetic code below, implements the casting rules
        # defined above.
        if isinstance(other, Quantity):
            return Quantity.with_dimensions(float(self) * float(other), self.dim * other.dim)
        elif is_scalar_type(other):
            return Quantity.with_dimensions(float(self) * other, self.dim)
        else:
            return NotImplemented
            #return super(Quantity,self).__mul__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        if isinstance(other, Quantity):
            return Quantity.with_dimensions(float(self) / float(other), self.dim / other.dim)
        elif is_scalar_type(other):
            return Quantity.with_dimensions(float(self) / other, self.dim)
        else:
            return NotImplemented
            #return super(Quantity,self).__div__(other)

    def __truediv__(self, other):
        if isinstance(other, Quantity):
            return Quantity.with_dimensions(float(self) / float(other), self.dim / other.dim)
        elif is_scalar_type(other):
            return Quantity.with_dimensions(float(self) / other, self.dim)
        else:
            return NotImplemented
            #return super(Quantity,self).__truediv__(other)

    def __rdiv__(self, other):
        if isinstance(other, Quantity):
            return Quantity.with_dimensions(float(other) / float(self), other.dim / self.dim)
        elif is_scalar_type(other):
            return Quantity.with_dimensions(other / float(self), [-x for x in self.dim._dims])
        else:
            return NotImplemented
            #return super(Quantity,self).__rdiv__(other)

    def __rtruediv__(self, other):
        if isinstance(other, Quantity):
            return Quantity.with_dimensions(float(other) / float(self), other.dim / self.dim)
        elif is_scalar_type(other):
            return Quantity.with_dimensions(other / float(self), [-x for x in self.dim._dims])
        else:
            return NotImplemented
            #return super(Quantity,self).__rtruediv__(other)

    def __mod__(self, other):
        if isinstance(other, Quantity) or is_scalar_type(other):
            dim = get_dimensions(other)
            if dim == self.dim:
                return Quantity.with_dimensions(float(self) % float(other), self.dim)
            else: raise DimensionMismatchError("Addition", self.dim, dim)
        else:
            return NotImplemented
            #return super(Quantity,self).__add__(other)    

    def __add__(self, other):
        if isinstance(other, Quantity) or is_scalar_type(other):
            dim = get_dimensions(other)
            if dim == self.dim:
                return Quantity.with_dimensions(float(self) + float(other), self.dim)
            else: raise DimensionMismatchError("Addition", self.dim, dim)
        else:
            return NotImplemented
            #return super(Quantity,self).__add__(other)    

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, Quantity) or is_scalar_type(other):
            dim = get_dimensions(other)
            if dim == self.dim:
                return Quantity.with_dimensions(float(self) - float(other), self.dim)
            else: raise DimensionMismatchError("Subtraction", self.dim, dim)
        else:
            return NotImplemented
            #return super(Quantity,self).__sub__(other)

    def __rsub__(self, other):
        if isinstance(other, Quantity) or is_scalar_type(other):
            dim = get_dimensions(other)
            if dim == self.dim:
                return Quantity.with_dimensions(float(other) - float(self), self.dim)
            else: raise DimensionMismatchError("Subtraction(R)", self.dim, dim)
        else:
            return NotImplemented
            #return super(Quantity,self).__rsub__(other)

    def __pow__(self, other):
        if isinstance(other, Quantity):
            if other.is_dimensionless():
                # WARNING: because dimension consistency is checked by exact comparison of dimensions,
                # this may lead to unexpected behaviour (e.g. (x**2)**0.5 may not have the same dimensions as x)
                return Quantity.with_dimensions(float(self) ** float(other), self.dim ** float(other))
            else: raise DimensionMismatchError("Power", self.dim, other.dim)
        elif is_scalar_type(other):
            return Quantity.with_dimensions(float(self) ** other, self.dim ** other)
        else:
            return NotImplemented
            #return super(Quantity,self).__pow__(other)

    def __rpow__(self, other):
        if self.is_dimensionless():
            if isinstance(other, Quantity):
                return Quantity.with_dimensions(float(other) ** float(self), other.dim ** float(self))
            elif is_scalar_type(other):
                return Quantity(other ** float(self))
            else:
                return NotImplemented
                #return super(Quantity,self).__pow__(other)
        else:
            raise DimensionMismatchError("Power(R)", self.dim)

    def __neg__(self):
        return Quantity.with_dimensions(-float(self), self.dim)

    def __pos__(self):
        return self

    def __abs__(self):
        return Quantity.with_dimensions(abs(float(self)), self.dim)
    #### COMPARISONS ####
    def __lt__(self, other):
        if isinstance(other, Quantity):
            if self.dim == other.dim:
                return float(self) < float(other)
            else: raise DimensionMismatchError("LessThan", self.dim, other.dim)
        elif is_scalar_type(other):
            if other == 0 or other == 0.: return float(self) < other
            if numpy.isposinf(other): return True
            if numpy.isneginf(other): return False
            if self.is_dimensionless():
                return float(self) < other
            else: raise DimensionMismatchError("LessThan", self.dim, Dimension())
        else:
            return NotImplemented
            #return super(Quantity,self).__lt__(other)

    def __le__(self, other):
        if isinstance(other, Quantity):
            if self.dim == other.dim:
                return float(self) <= float(other)
            else: raise DimensionMismatchError("LessThanOrEquals", self.dim, other.dim)
        elif is_scalar_type(other):
            if other == 0 or other == 0.: return float(self) <= other
            if numpy.isposinf(other): return True
            if numpy.isneginf(other): return False
            if self.is_dimensionless():
                return float(self) <= other
            else: raise DimensionMismatchError("LessThanOrEquals", self.dim, Dimension())
        else:
            return NotImplemented
            #return super(Quantity,self).__le__(other)

    def __gt__(self, other):
        if isinstance(other, Quantity):
            if self.dim == other.dim:
                return float(self) > float(other)
            else: raise DimensionMismatchError("GreaterThan", self.dim, other.dim)
        elif is_scalar_type(other):
            if other == 0 or other == 0.: return float(self) > other
            if numpy.isneginf(other): return True
            if numpy.isposinf(other): return False
            if self.is_dimensionless():
                return float(self) > other
            else: raise DimensionMismatchError("GreaterThan", self.dim, Dimension())
        else:
            return NotImplemented
            #return super(Quantity,self).__gt__(other)

    def __ge__(self, other):
        if isinstance(other, Quantity):
            if self.dim == other.dim:
                return float(self) >= float(other)
            else: raise DimensionMismatchError("GreaterThanOrEquals", self.dim, other.dim)
        elif is_scalar_type(other):
            if other == 0 or other == 0.: return float(self) >= other
            if numpy.isneginf(other): return True
            if numpy.isposinf(other): return False
            if self.is_dimensionless():
                return float(self) >= other
            else: raise DimensionMismatchError("GreaterThanOrEquals", self.dim, Dimension())
        else:
            return NotImplemented
            #return super(Quantity,self).__ge__(other)

    def __eq__(self, other):
        if isinstance(other, Quantity):
            if self.dim == other.dim:
                return float(self) == float(other)
            else: raise DimensionMismatchError("Equals", self.dim, other.dim)
        elif is_scalar_type(other):
            if other == 0 or other == 0. or numpy.isinf(other): return float(self) == other
            if self.dim.is_dimensionless():
                return float(self) == other
            else: raise DimensionMismatchError("Equals", self.dim, Dimension())
        else:
            return NotImplemented
            #return super(Quantity,self).__eq__(other)

    def __ne__(self, other):
        if isinstance(other, Quantity):
            if self.dim == other.dim:
                return float(self) != float(other)
            else: raise DimensionMismatchError("Equals", self.dim, other.dim)
        elif is_scalar_type(other):
            if other == 0 or other == 0. or numpy.isinf(other): return float(self) != other
            if self.dim.is_dimensionless():
                return float(self) != other
            else: raise DimensionMismatchError("NotEquals", self.dim, Dimension())
        else:
            return NotImplemented
            #return super(Quantity,self).__ne__(other)
    #### MAKE QUANTITY PICKABLE ####
    def __reduce__(self):
        return (quantity_with_dimensions, (float(self), self.dim))


class Unit(Quantity):
    '''
    A physical unit
    
    Normally, you do not need to worry about the implementation of
    units. They are derived from the :class:`Quantity` object with
    some additional information (name and string representation).
    You can define new units which will be used when generating
    string representations of quantities simply by doing an
    arithmetical operation with only units, for example::
    
        Nm = newton * metre
    
    Note that operations with units are slower than operations with
    :class:`Quantity` objects, so for efficiency if you do not need the
    extra information that a :class:`Unit` object carries around, write
    ``1*second`` in preference to ``second``.
    '''

    # original documentation
    """A physical unit
    
    Basically, a unit is just a quantity with given dimensions, e.g.
    mvolt = 0.001 with the dimensions of voltage. The units module
    defines a large number of standard units, and you can also define
    your own (see below).
    
    The unit class also keeps track of various things that were used
    to define it so as to generate a nice string representation of it.
    See Representation below.
    
    Typical usage:
    
    x = 3 * mvolt # returns a quantity
    print x.in_unit(uvolt) # returns 3000 uV 
    
    Standard units:
    
    The units class has the following fundamental units:
    
    metre, kilogram, second, amp, kelvin, mole, candle
    
    And these additional basic units:
    
    radian, steradian, hertz, newton, pascal, joule, watt,
    coulomb, volt, farad, ohm, siemens, weber, tesla, henry,
    celsius, lumen, lux, becquerel, gray, sievert, katal
    
    And additionally, it includes all scaled versions of these
    units using the following prefixes

     Factor     Name    Prefix
     -----      ----    ------
     10^24      yotta   Y
     10^21      zetta   Z
     10^18      exa     E
     10^15      peta    P
     10^12      tera    T
     10^9       giga    G
     10^6       mega    M
     10^3       kilo    k
     10^2       hecto   h
     10^1       deka    da
     1           
     10^-1      deci    d
     10^-2      centi   c
     10^-3      milli   m
     10^-6      micro   u (\mu in SI)
     10^-9      nano    n
     10^-12     pico    p
     10^-15     femto   f
     10^-18     atto    a
     10^-21     zepto   z
     10^-24     yocto   y
    
    So for example nohm, ytesla, etc. are all defined.
    
    Defining your own:
    
    It can be useful to define your own units for printing
    purposes. So for example, to define the newton metre, you
    write:
    
    Nm = newton * metre
    
    Writing:
    
    print (1*Nm).in_unit(Nm)
    
    will return "1 Nm" because the Unit class generates a new
    display name of "Nm" from the display names "N" and "m" for
    newtons and metres automatically (see Representation below).
    
    To register this unit for use in the automatic printing
    of the Quantity.in_best_unit() method, see the documentation
    for the UnitRegistry class.
    
    Construction:
    
    The best way to construct a new unit is to use standard units
    already defined and arithmetic operations, e.g. newton*metre.
    See the documentation for __init__ and the static methods create(...)
    and create_scaled_units(...) for more details.
    
    If you don't like the automatically generated display name for
    the unit, use the set_display_name(name) method.
    
    Representation:
    
    A new unit defined by multiplication, division or taking powers
    generates a name for the unit automatically, so that for
    example the name for pfarad/mmetre**2 is "pF/mm^2", etc. If you
    don't like the automatically generated name, use the 
    set_display_name(name) method.
    """
    __slots__ = ["dim", "scale", "scalefactor", "dispname", "name", "iscompound"]
    #### CONSTRUCTION ####
    def __init__(self, value):
        """Initialises a dimensionless unit
        """
        super(Unit, self).__init__(value)
        self.dim = Dimension()
        self.scale = [ "", "", "", "", "", "", "" ]
        self.scalefactor = ""
        self.dispname = ""
        self.iscompound = False

    def __new__(typ, *args, **kw):
        obj = super(Unit, typ).__new__(typ, *args, **kw)
        global automatically_register_units
        if automatically_register_units:
            register_new_unit(obj)
        return obj
    @staticmethod
    def create(dim, name="", dispname="", scalefactor="", **keywords):
        """Creates a new named unit
        
        dim -- the dimensions of the unit
        name -- the full name of the unit, e.g. volt
        dispname -- the display name, e.g. V
        scalefactor -- scaling factor, e.g. m for mvolt
        keywords -- the scaling for each SI dimension, e.g. length="m", mass="-1", etc.
        """
        scale = [ "", "", "", "", "", "", "" ]
        for k in keywords:
            scale[_di[k]] = keywords[k]
        v = 1.0
        for s, i in izip(scale, dim._dims):
            if i: v *= _siprefixes[s] ** i
        u = Unit(v * _siprefixes[scalefactor])
        u.dim = dim
        u.scale = scale
        u.scalefactor = scalefactor + ""
        u.name = name + ""
        u.dispname = dispname + ""
        u.iscompound = False
        return u
    @staticmethod
    def create_scaled_unit(baseunit, scalefactor):
        """Create a scaled unit from a base unit
        
        baseunit -- e.g. volt, amp
        scalefactor -- e.g. "m" for mvolt, mamp
        """
        u = Unit(float(baseunit) * _siprefixes[scalefactor])
        u.dim = baseunit.dim
        u.scale = baseunit.scale
        u.scalefactor = scalefactor
        u.name = scalefactor + baseunit.name
        u.dispname = scalefactor + baseunit.dispname
        u.iscompound = False
        return u
    #### METHODS ####
    def set_name(self, name):
        """Sets the name for the unit
        """
        self.name = name

    def set_display_name(self, name):
        """Sets the display name for the unit
        """
        self.dispname = name
    #### REPRESENTATION ####
    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self.dispname == "":
            s = self.scalefactor + " "
            for i in range(7):
                if self.dim._dims[i]:
                    s += self.scale[i] + _ilabel[i]
                    if self.dim._dims[i] != 1: s += "^" + str(self.dim._dims[i])
                    s += " "
            if not len(s): return "1"
            return s.strip()
        else:
            return self.dispname
    #### ARITHMETIC ####
    def __mul__(self, other):
        if isinstance(other, Unit):
            u = Unit(float(self) * float(other))
            u.name = self.name + other.name
            u.dispname = self.dispname + ' ' + other.dispname
            u.dim = self.dim * other.dim
            u.iscompound = True
            return u
        else:
            return super(Unit, self).__mul__(other)

    def __div__(self, other):
        if isinstance(other, Unit):
            u = Unit(float(self) / float(other))
            u.name = self.name + 'inv_' + other.name + '_endinv'
            if other.iscompound:
                u.dispname = '(' + self.dispname + ')'
            else:
                u.dispname = self.dispname
            u.dispname += '/'
            if other.iscompound:
                u.dispname += '(' + other.dispname + ')'
            else:
                u.dispname += other.dispname
            u.dim = self.dim / other.dim
            u.iscompound = True
            return u
        else:
            return super(Unit, self).__div__(other)

    def __pow__(self, other):
        if is_scalar_type(other):
            u = Unit(float(self) ** other)
            u.name = self.name + 'pow_' + str(other) + '_endpow'
            if self.iscompound:
                u.dispname = '(' + self.dispname + ')'
            else:
                u.dispname = self.dispname
            u.dispname += '^' + str(other)
            u.dim = self.dim ** other
            return u
        else:
            return super(Unit, self).__mul__(other)

automatically_register_units = False
#### FUNDAMENTAL UNITS
metre = Unit.create(Dimension(m=1), "metre", "m")
meter = Unit.create(Dimension(m=1), "meter", "m")
kilogram = Unit.create(Dimension(kg=1), "kilogram", "kg")
gram = Unit.create_scaled_unit(kilogram, "m")
gram.set_name('gram')
gram.set_display_name('g')
gramme = Unit.create_scaled_unit(kilogram, "m")
gramme.set_name('gramme')
gramme.set_display_name('g')
second = Unit.create(Dimension(s=1), "second", "s")
amp = Unit.create(Dimension(A=1), "amp", "A")
kelvin = Unit.create(Dimension(K=1), "kelvin", "K")
mole = Unit.create(Dimension(mol=1), "mole", "mol")
candle = Unit.create(Dimension(candle=1), "candle", "cd")
fundamental_units = [ metre, meter, gram, second, amp, kelvin, mole, candle ]

#### DERIVED UNITS, from http://physics.nist.gov/cuu/Units/units.html
derived_unit_table = \
        [\
        [ 'radian', 'rad', Dimension() ], \
        [ 'steradian', 'sr', Dimension() ], \
        [ 'hertz', 'Hz', Dimension(s= -1) ], \
        [ 'newton', 'N', Dimension(m=1, kg=1, s= -2) ], \
        [ 'pascal', 'Pa', Dimension(m= -1, kg=1, s= -2) ], \
        [ 'joule', 'J', Dimension(m=2, kg=1, s= -2) ], \
        [ 'watt', 'W', Dimension(m=2, kg=1, s= -3) ], \
        [ 'coulomb', 'C', Dimension(s=1, A=1) ], \
        [ 'volt', 'V', Dimension(m=2, kg=1, s= -3, A= -1) ], \
        [ 'farad', 'F', Dimension(m= -2, kg= -1, s=4, A=2) ], \
        [ 'ohm', 'ohm', Dimension(m=2, kg=1, s= -3, A= -2) ], \
        [ 'siemens', 'S', Dimension(m= -2, kg= -1, s=3, A=2) ], \
        [ 'weber', 'Wb', Dimension(m=2, kg=1, s= -2, A= -1) ], \
        [ 'tesla', 'T', Dimension(kg=1, s= -2, A= -1) ], \
        [ 'henry', 'H', Dimension(m=2, kg=1, s= -2, A= -2) ], \
        [ 'celsius', 'degC', Dimension(K=1) ], \
        [ 'lumen', 'lm', Dimension(cd=1) ], \
        [ 'lux', 'lx', Dimension(m= -2, cd=1) ], \
        [ 'becquerel', 'Bq', Dimension(s= -1) ], \
        [ 'gray', 'Gy', Dimension(m=2, s= -2) ], \
        [ 'sievert', 'Sv', Dimension(m=2, s= -2) ], \
        [ 'katal', 'kat', Dimension(s= -1, mol=1) ]\
        ]

# Pointless list only here so that static analysis in Eclipse works ok
# All the values here are overwritten by the code below
volt = Unit(1); mvolt = Unit(1); uvolt = Unit(1)
namp = Unit(1); mamp = Unit(1); uamp = Unit(1); pamp = Unit(1)
ohm = Unit(1); Mohm = Unit(1); kohm = Unit(1)
siemens = Unit(1); msiemens = Unit(1); usiemens = Unit(1)
hertz = Unit(1); khertz = Unit(1); Mhertz = Unit(1)
farad = Unit(1); mfarad = Unit(1); ufarad = Unit(1); nfarad = Unit(1)
msecond = Unit(1)

# Generate derived unit objects and make a table of base units from these and the fundamental ones
base_units = fundamental_units + [gramme, kilogram] # make a copy
for _du in derived_unit_table:
    _u = Unit.create(_du[2], _du[0], _du[1])
    exec _du[0] + "=_u"
    base_units.append(_u)

all_units = base_units + []

# Generate scaled units for all base units
scaled_units = []
for _bu in base_units:
    for _k in _siprefixes.keys():
        if len(_k):
            _u = Unit.create_scaled_unit(_bu, _k)
            exec _k + _bu.name + "=_u"
            all_units.append(_u)
            if not _k in ["da", "d", "c", "h"]:
                scaled_units.append(_u)

# Generate 2nd and 3rd powers for all scaled base units
powered_units = []
for bu in all_units + []:
    for i in [2, 3]:
        u = bu ** i
        u.name = bu.name + str(i)
        exec bu.name + str(i) + '=u'
        all_units.append(u)
        if not bu.scalefactor in ['da', 'd', 'c', 'h']:
            powered_units.append(u)

# Define additional units

# Current list from http://physics.nist.gov/cuu/Units/units.html, far from complete
additional_units = [ pascal * second, newton * metre, watt / metre ** 2, joule / kelvin, \
                   joule / (kilogram * kelvin), joule / kilogram, watt / (metre * kelvin), \
                   joule / metre ** 3, volt / metre ** 3, coulomb / metre ** 3, coulomb / metre ** 2, \
                   farad / metre, henry / metre, joule / mole, joule / (mole * kelvin), \
                   coulomb / kilogram, gray / second, katal / metre ** 3 ]

automatically_register_units = True


class UnitRegistry(object):
    """Stores known units for printing in best units
    
    All a user needs to do is to use the register_new_unit(u)
    function.
    
    Default registries:
    
    The units module defines three registries, the standard units,
    user units, and additional units. Finding best units is done
    by first checking standard, then user, then additional. New
    user units are added by using the register_new_unit(u) function.
    
    Standard units includes all the basic non-compound unit names
    built in to the module, including volt, amp, etc. Additional
    units defines some compound units like newton metre (Nm) etc.
    
    Methods:
    
    add(u) - add a new unit
    __getitem__(x) - get the best unit for quantity x
      e.g. UnitRegistry ur; ur[3*mvolt] returns mvolt
    """
    def __init__(self):
        self.objs = []

    def add(self, u):
        """Add a unit to the registry
        """
        self.objs.append(u)

    def __getitem__(self, x):
        """Returns the best unit for quantity x
        
        The algorithm is to consider the value:
        
        m=abs(x/u)
        
        for all matching units u. If there is a unit u with a value of
        m in [1,1000) then we select that unit. Otherwise, we select
        the first matching unit (which will typically be the unscaled
        version).
        """
        matching = filter(lambda o: have_same_dimensions(o, x), self.objs)
        if len(matching) == 0:
            raise KeyError("Unit not found in registry.")
        floatrep = filter(lambda o: 0.1 <= abs(float(x / o)) < 100, matching)
        if len(floatrep):
            return floatrep[0]
        else:
            return matching[0]

def register_new_unit(u):
    """Register a new unit for automatic displaying of quantities
    
    Example usage:
    
    2.0*farad/metre**2 = 2.0 m^-4 kg^-1 s^4 A^2
    register_new_unit(pfarad / mmetre**2)
    2.0*farad/metre**2 = 2000000.0 pF/mm^2
    """
    UserUnitRegister.add(u)

standard_unit_register = UnitRegistry()
additional_unit_register = UnitRegistry()
UserUnitRegister = UnitRegistry()
map(standard_unit_register.add, base_units + scaled_units + powered_units)
map(additional_unit_register.add, additional_units)

def all_registered_units(*regs):
    """Returns all registered units in the correct order
    """
    if not len(regs):
        regs = [ standard_unit_register, UserUnitRegister, additional_unit_register]
    for r in regs:
        for u in r.objs:
            yield u

def _get_best_unit(x, *regs):
    """Returns the best unit for quantity x
    
    Checks the registries regs, unless none are provided in which
    case it will check the standard, user and additional unit
    registers in turn.
    """
    if get_dimensions(x) == Dimension():
        return Quantity(1)
    if len(regs):
        for r in regs:
            try:
                return r[x]
            except KeyError:
                pass
        return Quantity.with_dimensions(1, x.dim)
    else:
        return _get_best_unit(x, standard_unit_register, UserUnitRegister, additional_unit_register)

def get_unit(x, *regs):
    '''
    Find the most appropriate consistent unit from the unit registries, or just return a Quantity with the same dimensions and value 1
    '''
    for u in all_registered_units(*regs):
        if is_equal(float(u), 1) and have_same_dimensions(u, x):
            return u
    return Quantity.with_dimensions(1, get_dimensions(x))

def get_unit_fast(x):
    '''
    Return a quantity with value 1 and the same dimensions
    '''
    return Quantity.with_dimensions(1, get_dimensions(x))

#### DECORATORS


def check_units(**au):
    """Decorator to check units of arguments passed to a function
    
    **Sample usage:** ::
    
        @check_units(I=amp,R=ohm,wibble=metre,result=volt)
        def getvoltage(I,R,**k):
            return I*R

    You don't have to check the units of every variable in the function, and
    you can define what the units should be for variables that aren't
    explicitly named in the definition of the function. For example, the code
    above checks that the variable wibble should be a length, so writing::
    
        getvoltage(1*amp,1*ohm,wibble=1)
    
    would fail, but::
    
        getvoltage(1*amp,1*ohm,wibble=1*metre)
    
    would pass.
    String arguments are not checked (e.g. ``getvoltage(wibble='hello')`` would pass).
    
    The special name ``result`` is for the return value of the function.
    
    An error in the input value raises a :exc:`DimensionMismatchError`, and an error
    in the return value raises an ``AssertionError`` (because it is a code
    problem rather than a value problem).
    
    **Notes**
    
    This decorator will destroy the signature of the original function, and
    replace it with the signature ``(*args, **kwds)``. Other decorators will
    do the same thing, and this decorator critically needs to know the signature
    of the function it is acting on, so it is important that it is the first
    decorator to act on a function. It cannot be used in combination with another
    decorator that also needs to know the signature of the function.
    """
    def do_check_units(f):
        @wraps(f)
        def new_f(*args, **kwds):
            newkeyset = kwds.copy()
            arg_names = f.func_code.co_varnames[0:f.func_code.co_argcount]
            for (n, v) in zip(arg_names, args[0:f.func_code.co_argcount]):
                newkeyset[n] = v
            for k in newkeyset.iterkeys():
                if (k in au.keys()) and not isinstance(newkeyset[k], str): # string variables are allowed to pass, the presumption is they name another variable
                    if not have_same_dimensions(newkeyset[k], au[k]):
                        raise DimensionMismatchError("Function " + f.__name__ + " variable " + k + " should have dimensions of " + str(au[k]), get_dimensions(newkeyset[k]))
            result = f(*args, **kwds)
            if "result" in au:
                assert have_same_dimensions(result, au["result"]), \
                    "Function " + f.__name__ + " should return a value with unit " + str(au["result"]) + " but has returned " + str(get_dimensions(result))
            return result
#        new_f.__name__ = f.__name__
#        new_f.__doc__ = f.__doc__
#        new_f.__dict__.update(f.__dict__)
        return new_f
    return do_check_units

## Note: do not normally call this, see note on importing of decorator module at the top of this module
#if use_decorator:
#    old_check_units = check_units
#    def check_units(**au):
#        return lambda f : decorator.new_wrapper(old_check_units(**au)(f), f)
#    check_units.__doc__ = old_check_units.__doc__

def _check_nounits(**au):
    """Don't bother checking units decorator
    """
    def dont_check_units(f):
        return f
    return dont_check_units

def scalar_representation(x):
    if isinstance(x, Unit):
        return x.name
    u = get_unit(x)
    if isinstance(u, Unit):
        return '(' + repr(float(x)) + '*' + u.name + ')'
    if isinstance(x, Quantity):
        return '(Quantity.with_dimensions(' + repr(float(x)) + ',' + repr(x.dim._dims) + '))'
    return repr(x)

# Remove all units
if not bup.use_units:
    for _u in all_units:
        exec _u.name + "=float(_u)"
    check_units = _check_nounits
    def get_dimensions(obj):
        return Dimension()

    def is_dimensionless(obj):
        return True

    def have_same_dimensions(obj1, obj2):
        return True

    def get_unit(x, *regs):
        return 1.

    def scalar_representation(x):
        return '1.0'

# Add unit names to __all__
all_unit_names = [u.name for u in all_units]
__all__.extend(all_unit_names)

if __name__ == "__main__":

#    # the pattern 'pat' below is a regular expression for all the unit names
#    # you can use it as an exclusion pattern for the epydoc api docs
#    base_unit_ids = set([id(_) for _ in base_units])
#    l = [_ for _ in __all__ if id(locals()[_]) in base_unit_ids]
#    anybaseunit = '('+'|'.join(l)+')'
#    print anybaseunit
#    prefixes = [_ for _ in _siprefixes.keys() if _]
#    anyprefix = '('+'|'.join(prefixes)+')'
#    print anyprefix
#    pat = anyprefix+'?'+anybaseunit+'[23]?'
#    print pat
#    import re
#    for x in __all__:
#        if not len(re.findall(pat,x)):
#            print x

    from numpy import *

    # shorthand function used for example code below
    def pE(vname, str):
        uname = vname
        if vname == "": uname = "temp"
        exec(uname + "=" + str)
        if vname != "": print vname, "=",
        print str,
        if locals()[uname] != None:
            print '=', locals()[uname]
        else:
            print
        return locals()[uname]

    V = pE("V", "3 * volt")
    I = pE("I", "2 * amp")
    a = pE("a", "array([1,2,3])")
    print
    R = pE("R", "V/I")
    pE("", "I*R")
    print
    pE("", "a*V")
    pE("", "V*a")
    pE("", "a+V")
    print
    pE("", "1000*metre")
    pE("", "1000*mmetre")
    print
    pE("", "(2*volt).in_unit(mvolt)")
    pE("", "(2*volt)/mvolt")
    print "(2*volt).in_unit(amp) =",
    try:
        print (2 * volt).in_unit(amp)
    except DimensionMismatchError, i:
        print "DimensionMismatchError:", i
    print
    pE("", "have_same_dimensions(1*volt,1*amp*ohm)")
    pE("", "have_same_dimensions(1*volt*second,1*amp*ohm)")
    print
    pE("", "(ufarad/nmetre)**2")
    print
    pE("", "2.0*farad/metre**2")
    pE("", "register_new_unit(pfarad / mmetre**2)")
    pE("", "2.0*farad/metre**2")

    print
    print "Some decorator examples (see code):"
    print

    @check_units(I=amp, R=ohm, wibble=metre, result=volt)
    def getvoltage(I, R, *args, **k):
        return I * R

    try:
        print getvoltage(1 * amp, 2 * ohm, 20)
        print getvoltage(R=2 * ohm, I=1 * amp, wibble=7 * mmetre)
        print getvoltage(1 * amp, 2 * ohm * metre)
    except DimensionMismatchError, inst:
        print "DME:", inst

    print
    pE("", "get_unit(3*msecond)")

###################################################
##### ADDITIONAL INFORMATION

#SI DIMENSIONS
#-------------
#Quantity               Unit      Symbol
#--------               ----      ------
#Length                 metre     m
#Mass                   kilogram  kg
#Time                   second    s
#Electric current       ampere    A
#Temperature            kelvin    K
#Quantity of substance  mole      mol
#Luminosity             candle    cd

# SI UNIT PREFIXES
# ----------------
# Factor     Name    Prefix
# -----      ----    ------
# 10^24      yotta   Y
# 10^21      zetta   Z
# 10^18      exa     E
# 10^15      peta    P
# 10^12      tera    T
# 10^9       giga    G
# 10^6       mega    M
# 10^3       kilo    k
# 10^2       hecto   h
# 10^1       deka    da
# 1           
# 10^-1      deci    d
# 10^-2      centi   c
# 10^-3      milli   m
# 10^-6      micro   u (\mu in SI)
# 10^-9      nano    n
# 10^-12     pico    p
# 10^-15     femto   f
# 10^-18     atto    a
# 10^-21     zepto   z
# 10^-24     yocto   y



"""
Setup script for Brian
"""

from distutils.core import setup

version = '1.3.1'

long_description = '''
Brian is a simulator for spiking neural networks available on almost all platforms.
The motivation for this project is that a simulator should not only save the time of
processors, but also the time of scientists.

Brian is easy to learn and use, highly flexible and easily extensible. The Brian package
itself and simulations using it are all written in the Python programming language,
which is an easy, concise and highly developed language with many advanced features and
development tools, excellent documentation and a large community of users providing
support and extension packages.
'''

# the create_extras.py script will automatically generate an extras files
# containing the following files
extras_folders = ['tutorials/tutorial1_basic_concepts/*.py',
                    'tutorials/tutorial1_basic_concepts/*.txt',
                  'tutorials/tutorial2_connections/*.py',
                    'tutorials/tutorial2_connections/*.txt',
                  'examples/*.py', 'examples/*.txt',
                    'examples/audition/*.py',
                    'examples/electrophysiology/*.py',
                    'examples/electrophysiology/*.npy',
                    'examples/frompapers/*.py',
                    'examples/hears/*.py',
                    'examples/interface/*.*',
                    'examples/misc/*.py',
                    'examples/modelfitting/*.py',
                    'examples/modelfitting/*.txt',
                    'examples/multiprocessing/*.py',
                    'examples/plasticity/*.py',
                  'docs/*.*',
                    'docs/_images/*.jpg',
                    'docs/_sources/*.*',
                    'docs/_static/*.*' 
                  ]

if __name__ == "__main__":
    setup(name='brian',
      version=version,
      py_modules=['brian_unit_prefs', 'brian_no_units', 'brian_no_units_no_warnings'],
      packages=['brian',
                    'brian.connections',
                    'brian.deprecated',
                    'brian.experimental',
                        'brian.experimental.cuda',
                        'brian.experimental.codegen',
                        'brian.experimental.compensation',
                        'brian.experimental.modelfitting',
                        'brian.experimental.neuromorphic',
                        'brian.experimental.stdp',
                    'brian.hears',
                        'brian.hears.filtering',
                        'brian.hears.hrtf',
                    'brian.library',
                        'brian.library.electrophysiology',
                        'brian.library.modelfitting',
                    'brian.tests',
                        'brian.tests.testcorrectness',
                        'brian.tests.testinterface',
                        'brian.tests.testutils',
                        'brian.tests.testfeatures',
                    'brian.tools',
                    'brian.utils',
                        'brian.utils.ccircular',
                        'brian.utils.fastexp',
                        'brian.utils.sparse_patch',
                ],
      package_data={'brian.utils.ccircular':['*.cxx', '*.h', '*.i', '*.cpp', '*.bat'],
                    'brian.utils.fastexp':['*.cxx', '*.h', '*.i', '*.cpp', '*.bat', '*.c']},
      requires=['matplotlib(>=0.90.1)',
                'numpy(>=1.3.0)',
                'scipy(>=0.7.0)'
                ],
      url='http://www.briansimulator.org/',
      description='A clock-driven simulator for spiking neural networks',
      long_description=long_description,
      author='Romain Brette, Dan Goodman',
      author_email='Romain.Brette at ens.fr',
      download_url='https://neuralensemble.org/trac/brian/wiki/Downloads',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: Other/Proprietary License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
        ]
      )



from brian import *
from brian.utils.separate_equations import *

eqs = Equations('''
dx/dt = -x/second : 1
dy/dt = -y/second : 1
z     = x-y       : 1
du/dt = -u/second : 1
dv/dt = -v/second : 1
w     = u-v       : 1
''')

print separate_equations(eqs)

eqs = Equations('''
dx/dt = -x/second : 1
y = x*x : 1
z = x
du/dt = -u/second : 1
#v = z+u : 1
''')

print separate_equations(eqs)



from brian import *

__all__ = ['get_samplerate', 'set_default_samplerate']

default_samplerate = 44.1*kHz

def get_samplerate(samplerate):
    if samplerate is None:
        return default_samplerate
    else:
        return samplerate
    
def set_default_samplerate(samplerate):
    '''
    Sets the default samplerate for Brian hears objects, by default 44.1 kHz.
    '''
    global default_samplerate
    default_samplerate = samplerate



'''
Realtime plotting example
'''

# These lines are necessary for interactive plotting when launching from the
# Eclipse IDE, they may not be necessary in every environment.
import matplotlib
matplotlib.use('WXAgg') # You may need to experiment, try WXAgg, GTKAgg, QTAgg, TkAgg

from brian import *
###### Set up the standard CUBA example ######
N = 4000
eqs = '''
dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
dge/dt = -ge/(5*ms) : volt
dgi/dt = -gi/(10*ms) : volt
'''
P = NeuronGroup(N, eqs, threshold= -50 * mV, reset= -60 * mV)
P.v = -60 * mV + 10 * mV * rand(len(P))
Pe = P.subgroup(3200)
Pi = P.subgroup(800)
Ce = Connection(Pe, P, 'ge', weight=1.62 * mV, sparseness=0.02)
Ci = Connection(Pi, P, 'gi', weight= -9 * mV, sparseness=0.02)

M = SpikeMonitor(P)
trace = RecentStateMonitor(P, 'v', record=range(5), duration=200 * ms)

ion()
subplot(211)
raster_plot(M, refresh=10 * ms, showlast=200 * ms, redraw=False)
subplot(212)
trace.plot(refresh=10 * ms, showlast=200 * ms)

run(1 * second)

ioff() # switch interactive mode off
show() # and wait for user to close the window before shutting down



# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
"""Util to test if two floating point numbers are equal

Use these functions for very precise tests of equality:

-- is_equal(x,y), for x=y
-- is_less_than_or_equal(x,y), for x<=y
-- is_greater_than_or_equal(x,y), for x>=y

Use these functions for less precise tests (for example, if you have done some operations on two varaibles):

-- is_approx_equal(x,y), for x=y
-- is_approx_less_than_or_equal(x,y), for x<=y
-- is_approx_greater_than_or_equal(x,y), for x>=y

The underlying mechanism is that the more precise version tests for equality using machine epsilon
precision, that is, x=y if abs(x-y)<abs(x)*epsilon where epsilon is the smallest value such that
1+epsilon>epsilon. The less precise mechanism simply uses 100*epsilon instead of epsilon.

Use this function for testing if you want to specify an absolute tolerance:

-- is_within_absolute_tolerance(x,y[,absolutetolerance])
    The default tolerance is the sqrt of epsilon, or about 1e-8 for a 64 bit float

Note also that you can use the numpy function:

-- allclose(a, b, rtol = 1e-5, atol = 1e-8)
    Where rtol is the relative tolerance, and atol is the absolute tolerance which comes into
    play when the numbers are very close to zero.

Warning: none of these functions can be guaranteed to work in the way you might
expect them to. Errors can accumulate to the point where even 100*epsilon is an inappropriate test
for approximate equality.
"""

import math

# This finds the 'machine epsilon' for the current hardware float type, the
# smallest value of epsilon so that 1+epsilon>1
epsilon = 1.
while 1. + epsilon > 1.:
    epsilon /= 2
epsilon *= 2.
# Result for 32 bit float should be: 1.1929093e-7 
# For 64 bit float should be:        2.220446049250313e-16

# This value can be used for more approximate testing
approxepsilon = epsilon * 10000

# This value is the default tolerance for medium sized numbers (used in the units class)
defaultabsolutetolerance = math.sqrt(epsilon) # 1.4901161193847656e-008 on 64 bit system

def is_equal(x, y):
    if x == y: return True
    return abs(x - y) < abs(x) * epsilon

def is_less_than_or_equal(x, y):
    return x < y or is_equal(x, y)

def is_greater_than_or_equal(x, y):
    return x > y or is_equal(x, y)

def is_approx_equal(x, y):
    if x == y: return True
    return abs(x - y) < abs(x) * approxepsilon

def is_approx_less_than_or_equal(x, y):
    return x < y or is_approx_equal(x, y)

def is_approx_greater_than_or_equal(x, y):
    return x > y or is_approx_equal(x, y)

def is_within_absolute_tolerance(x, y, absolutetolerance=defaultabsolutetolerance):
    return float(abs(x - y)) < absolutetolerance



'''
Example of basic filtering of a sound with Brian hears.
This example implements a cochleagram based on a gammatone filterbank
followed by halfwave rectification, cube root compression and 10 Hz
low pass filtering.
'''
from brian import *
from brian.hears import *

sound1 = tone(1*kHz, .1*second)
sound2 = whitenoise(.1*second)

sound = sound1+sound2
sound = sound.ramp()

cf = erbspace(20*Hz, 20*kHz, 3000)
gammatone = Gammatone(sound, cf)
cochlea = FunctionFilterbank(gammatone, lambda x: clip(x, 0, Inf)**(1.0/3.0))
lowpass = LowPass(cochlea, 10*Hz)
output = lowpass.process()

imshow(output.T, origin='lower left', aspect='auto', vmin=0)
show()



from brian import *
import time

client = RemoteControlClient()

def compute_rate(spikes):
    i, t = zip(*spikes)
    i = array(i)
    t = array(t)
    tmin = amin(t)
    tmax = amax(t)
    return len(i) / (tmax - tmin)

spikes = client.evaluate('M.spikes')
print 'Rate:', compute_rate(spikes)
client.execute('Ce.W.alldata[:]*=2')
time.sleep(5)
spikes = client.evaluate('M.spikes')
print 'Rate:', compute_rate(spikes)
client.execute('stop()')



import dev.ideas.synapses.statevectors as sv
import numpy as np

reload(sv)

v = sv.ConstructionSparseStateVector(4, (np.int8, np.int8, np.float64, np.float64))

nvalues = 52

values = []
values.append(np.arange(nvalues, dtype = np.int8))
values.append(np.arange(nvalues, dtype = np.int8)[::-1])
values.append(np.array(np.random.rand(nvalues), dtype = np.float64))
values.append(np.array(np.random.rand(nvalues), dtype = np.float64))

v.append(values)

v.append(values)

v.append(values)

vcomp = v.compress()




'''
Adaptive exponential integrate-and-fire model.
http://www.scholarpedia.org/article/Adaptive_exponential_integrate-and-fire_model

Introduced in Brette R. and Gerstner W. (2005), Adaptive Exponential
Integrate-and-Fire Model as an Effective Description of Neuronal Activity,
J. Neurophysiol. 94: 3637 - 3642. 
'''
from brian import *

# Parameters
C = 281 * pF
gL = 30 * nS
taum = C / gL
EL = -70.6 * mV
VT = -50.4 * mV
DeltaT = 2 * mV
Vcut = VT + 5 * DeltaT

# Pick an electrophysiological behaviour
tauw, a, b, Vr = 144 * ms, 4 * nS, 0.0805 * nA, -70.6 * mV # Regular spiking (as in the paper)
#tauw,a,b,Vr=20*ms,4*nS,0.5*nA,VT+5*mV # Bursting
#tauw,a,b,Vr=144*ms,2*C/(144*ms),0*nA,-70.6*mV # Fast spiking

eqs = """
dvm/dt=(gL*(EL-vm)+gL*DeltaT*exp((vm-VT)/DeltaT)+I-w)/C : volt
dw/dt=(a*(vm-EL)-w)/tauw : amp
I : amp
"""

neuron = NeuronGroup(1, model=eqs, threshold=Vcut, reset="vm=Vr;w+=b", freeze=True)
neuron.vm = EL
trace = StateMonitor(neuron, 'vm', record=0)
spikes = SpikeMonitor(neuron)

run(20 * ms)
neuron.I = 1 * nA
run(100 * ms)
neuron.I = 0 * nA
run(20 * ms)

# We draw nicer spikes
vm = trace[0]
for _, t in spikes.spikes:
    i = int(t / defaultclock.dt)
    vm[i] = 20 * mV

plot(trace.times / ms, vm / mV)
show()



"""
Dynamics of a network of sparsely connected inhibitory current-based 
integrate-and-fire neurons. Individual neurons fire irregularly at 
low rate but the network is in an oscillatory global activity regime 
where neurons are weakly synchronized.

Reference:
    "Fast Global Oscillations in Networks of Integrate-and-Fire
    Neurons with Low Firing Rates"
    Nicolas Brunel & Vincent Hakim
    Neural Computation 11, 1621-1671 (1999)
"""

from brian import *

N = 5000
Vr = 10 * mV
theta = 20 * mV
tau = 20 * ms
delta = 2 * ms
taurefr = 2 * ms
duration = .1 * second
C = 1000
sparseness = float(C)/N
J = .1 * mV
muext = 25 * mV
sigmaext = 1 * mV

eqs = """
dV/dt = (-V+muext + sigmaext * sqrt(tau) * xi)/tau : volt
"""

group = NeuronGroup(N, eqs, threshold=theta,
                    reset=Vr, refractory=taurefr)
group.V = Vr
conn = Connection(group, group, state='V', delay=delta,
                  weight = -J,
                  sparseness=sparseness)
M = SpikeMonitor(group)
LFP = PopulationRateMonitor(group, bin=0.4 * ms)

run(duration)

subplot(211)
raster_plot(M)
xlim(0, duration/ms)

subplot(212)
plot(LFP.times_/ms, LFP.rate)
xlim(0, duration/ms)

show()



"""Generate re_structured_text data for the tutorials

Reads Brian/Examples/Tutorials/tutorial*.py and generates the file
briantutorials.txt as restructuredtext for inclusion into the main
documentation file brian.txt. This file assumes the tutorial files
are organised as follows:

'''
Text explaining the next bit of code
'''

code

'''
More text
'''

more code

etc.

The text fragments are just inserted directly into the file, and
so we should use the following sectioning standard:

Tutorial title
**************

Tutorial section
~~~~~~~~~~~~~~~~

The code fragments are reformatted according to the restructedtext
format, which is simply that you write

::

    Code
    Lines
    Here

(i.e. :: followed by a tab indented block of code).

"""

import glob # generates directory listing
import os

os.chdir('../../../docs_sphinx') # work from docs_sphinx/ directory until we make the big change

# this takes a file name and returns a list of strings representing the lines
# to be output to the restructuredtext file, and the title of the tutorial
def generate_tutorial(fname):
    title = fname
    f = open(fname)
    lines = f.readlines()
    foundtitle = False
    # we use mode='code' and mode='text' to switch between processing text to be
    # inserted directly (which corresponds to the ''' comment ''' text in the
    # tutorial, and code which needs to be reformatted
    mode = 'code'
    tut = [] # a list of pairs (mode,lines)
    curlines = [] # the current list of lines to be added to the next (mode,lines) pair
    for l in lines:
        if l[0:3] == "'''" or l[0:3] == '"""': # a mode switch
            if len(curlines): # if there are lines to write then write them
                tut.append((mode, curlines))
            curlines = []
            if mode == 'code':
                mode = 'text'
            elif mode == 'text':
                mode = 'code'
        else:
            if not foundtitle:
                title = l # Assumption: first line is the title
                foundtitle = True
            curlines.append(l)
    # make sure we've got the last block
    if len(curlines):
        tut.append((mode, curlines))
    tutlines = [] # the output of the function, lines of restructuredtext
    for mode, lines in tut:
        if mode == 'code':
            # reformat by adding a paragraph with just :: before it
            tutlines.append('\n')
#            tutlines.append('.. class:: tutorialcode\n')
#            tutlines.append('..\n\n')
            tutlines.append('  ::\n\n')
            for l in lines:
                if len(l.strip()):
                    #tutlines.append('    :literal:`'+l[:-1]+'`\n') # add a tab of 4 spaces before each code line and mark it literal
                    tutlines.append('    ' + l[:-1] + '\n') # add a tab of 4 spaces before each code line and mark it literal
                else:
                    tutlines.append('    \n') # add a tab of 4 spaces before each code line
            tutlines.append('\n')
        if mode == 'text':
            tutlines.extend(lines)
    return (tutlines, title)

# these are all the tutorial files, when there are more than 9 of them we'll need to add
# an ordering here
tutorial_files = []
tutorial_series = []
for roots, dirs, files in os.walk('../tutorials'):
    newdirs = [_ for _ in dirs if '.svn' not in _]
    if '.svn' not in roots and len(newdirs):
        tutorial_series = newdirs
    if 'tutorial' in roots and '.svn' not in roots:
        for f in files:
            if 'tutorial' in f and '.py' in f:
                relpath = os.path.normpath(os.path.join(roots, f))
                tutrelpath = relpath
                tutrelpath = tutrelpath.replace('../tutorials/', '')
                tutrelpath = tutrelpath.replace('..\\tutorials\\', '')
                if '/' in tutrelpath or '\\' in tutrelpath:
                    tutorial_files.append(relpath)

for seriesname in tutorial_series:
    tutdesc = open('../tutorials/' + seriesname + '/introduction.txt', 'r').read()
    seriesfile = open(seriesname + '.txt', 'w')
    seriesfile.write('.. currentmodule:: brian\n\n')
    seriesfile.write('.. _' + seriesname + ':\n\n')
    seriesfile.write(tutdesc)
    seriesfile.write('\n\n**Tutorial contents**\n\n')
    seriesfile.write('.. toctree::\n')
    seriesfile.write('   :maxdepth: 2\n\n')
    for fname, i in zip(tutorial_files, range(1, 1 + len(tutorial_files))):
        if seriesname in fname:
            tutlines, tutorial_title = generate_tutorial(fname)
            # assumption: case is significant and tutorials are in a subfolder of tutorials with name including tutorials only once
            # e.g. tutorials/tutorial1-basic-concepts/tutorial1a-basic-concepts.py
            head, tail = os.path.split(fname)
            tutorial_name = tail[:tail.index('.py')]
            spaces = 5 - len(str(i))
            curtut_txt = open(tutorial_name + '.txt', 'w')
            curtut_txt.write('.. currentmodule:: brian\n\n')
            for line in tutlines:
                curtut_txt.write(line)
            del curtut_txt
            seriesfile.write('   ' + tutorial_name + '\n')
    seriesfile.close()



import brian_no_units
from brian import *
import numpy
import brian
import brian.experimental.modelfitting as gpu
import fast_compensation as cpu
#brian.log_level_error()
debug_level()

def compensate_gpu(input, trace, p=1.0):
    
    equations = Equations('''
            dV0/dt=(R*Iinj-V0+Vr)/tau : 1
            Iinj=(V-V0)/Re : 1
            dV/dt=Re*(I-Iinj)/taue : 1
            Ve=V-V0 : 1
            I : 1
            R : 1
            Re : 1
            Vr : 1
            tau : second
            taue : second
        ''')
    
    params = dict(      R =  [1.0e3, 1.0e6, 1000.0e6, 1.0e12],
                        tau =  [.1*ms, 1*ms, 30*ms, 200*ms],
                        Vr = [-100.0e-3, -80e-3, -40e-3, -10.0e-3],
                        Re = [1.0e3, 1.0e6, 1000.0e6, 1.0e12],
                        taue = [.01*ms, .1*ms, 5*ms, 20*ms])
   
   
   
    tracecomp, traceV, traceVe, results = gpu.compensate(input,
                                                          trace,
                                                          gpu=1,
                                                          p=p,
                                                          equations=equations,
                                                          maxiter=10,
                                                          popsize=10,
                                                          slice_duration=1*second,
                                                          **params)
    
    return traceV, traceVe, results






if __name__ == '__main__':

    I = numpy.load("current1.npy")[:10000] * 25 # because current1[2010_12_09_0006] * 25
    Vraw = numpy.load("trace1.npy")[:10000]

    R = 600*Mohm
    tau = 10*ms
    Vr = -70*mV
    Re = 100*Mohm
    taue = 1*ms

    # CPU COMPENSATION
    #Vcomp, params = cpu.compensate(I, Vraw, dt=.1*ms,
    #                                p = 1.0,
    #                                durslice=10*second,
    #                                R=R, tau=tau, Vr=Vr, Re=Re, taue=taue)

    # GPU COMPENSATION
    traceV, traceVe, results = compensate_gpu(I, Vraw)
    Vcomp_gpu = Vraw - traceVe

    subplot(211)
    plot(I)

    subplot(212)
    plot(Vraw)
    plot(Vcomp_gpu)

    show()



# Delays arising from interaction of synaptic conductances
# The idea: changing ge changes the delay
# Second idea: synaptic depression/facilitation changes the delay too
from brian import *

taum = 20 * ms
Ee = 70 * mV # relative to rest;
Ei = 0 * mV # shunting inhibition
taue = 5 * ms
taui = 10 * ms
N0 = 10
N = N0 * N0

eqs = '''
dv/dt=(-v+ge*(Ee-v)+gi*(Ei-v))/taum :  volt
dge/dt=-ge/taue : 1 # relative conductance
dgi/dt=-gi/taui : 1 # relative conductance
'''

neurons = NeuronGroup(N, model=eqs)
ge_range = linspace(0, 5, N0)
gi_range = linspace(0, 5, N0)
neurons.ge = array(ge_range * ones((N0, N0))).reshape(N)
neurons.gi = array(gi_range * ones((N0, N0))).T.reshape(N)

mon = StateMonitor(neurons, 'v', record=True) # that's heavy!

run(100 * ms)

m = zeros(N)
for i in range(N):
    m[i] = mon.times[argmax(mon[i])] / ms

figure()
hist(m, 10)
figure()
pcolor(array(m).reshape(N0, N0))
#contour(array(m).reshape(N0,N0))
xlabel('ge')
ylabel('gi')
figure()
plot(mon.times / ms, mon[2 * N0 + 2] / mV)
show()
